"""Test the higher level behavior of the unit commitment"""
import random
from minpower.generators import Generator
import pandas as pd
from pandas.testing import assert_series_equal
from .test_utils import *


@istest
@with_setup(setup=get_duals, teardown=reset_config)
def prices():
    """
    Run a basic unit commitment.
    Ensure that the correct LMPs are returned for all times.
    """
    generators = [make_cheap_gen(pmax=100), make_mid_gen(pmax=20), make_expensive_gen()]

    power_system, times = solve_problem(
        generators, do_reset_config=False, **make_loads_times(Pdt=[80, 110, 130])
    )
    lmps = [power_system.buses[0].price(t) for t in times]

    assert lmps == [gen_costs["cheap"], gen_costs["mid"], gen_costs["expensive"]]


@istest
def rolling():
    """
    Run a basic unit commitment over 72hrs
    Ensure that the generation meets the load for each time.
    """
    generators = [Generator(costcurveequation="10P+.01P^2")]
    Pdt = [random.randrange(0, 200) for i in range(0, 72)]
    power_system, times = solve_problem(generators, **make_loads_times(Pdt=Pdt))
    load = power_system.loads()[0]
    load_balanced = all(value(generators[0].power(t)) == load.power(t) for t in times)
    assert load_balanced


@istest
@with_setup(get_duals, reset_config)
def load_shedding():
    """
    Create a single generator with a high limit.
    Create a load that exceeds that limit at t1.
    Ensure that:
    * Pdt1=Pgmax
    * Pshedt1 = Pdt1 - pmax
    * pricet1 = cost of load shedding
    """
    pmax = 100
    Pdt1 = 211
    generators = [make_cheap_gen(pmax=pmax)]
    Pdt = [110, Pdt1, 110]
    power_system, times = solve_problem(
        generators, do_reset_config=False, **make_loads_times(Pdt=Pdt)
    )
    load = power_system.loads()[0]
    load_t1 = load.power(times[1], evaluate=True)
    load_t1_shed = load.shed(times[1], evaluate=True)
    price_t1 = power_system.buses[0].price(times[1])
    assert load_t1 == pmax
    assert load_t1_shed == Pdt1 - pmax
    assert price_t1 == user_config.cost_load_shedding


@istest
@with_setup(teardown=reset_config)
def reserve_fixed_amount():
    """
    Create two generators, the cheaper one with a limit near the load.
    Require 50MW of reserve (which exceeds pmax of cheap)
    Ensure that: g2 is turned on to meet reserve req.

    Pg1+Pg2 = load
    Pavail1 + Pavail2 = load + reserve
    """
    reserve_req = 30.0
    user_config.reserve_fixed = reserve_req

    generators = [
        make_cheap_gen(pmax=100),
        make_expensive_gen(pmin=0, pmax=300, costcurveequation="5000+30P"),
    ]
    Pdt = [80, 90]

    power_system, times = solve_problem(generators, **make_loads_times(Pdt=Pdt))
    total_power = [
        sum([value(gen.power(time)) for gen in generators]) for time in times
    ]
    total_available_power = [
        sum(value(gen.power_available(time)) for gen in generators) for time in times
    ]

    #    print load_t1 + reserve_req
    #    print [[value(gen.power(time)) for gen in generators] for time in times]
    #    print total_available_power
    #    print [reserve_req + Pd for Pd in Pdt]

    assert Pdt == total_power
    assert total_available_power >= [reserve_req + Pd for Pd in Pdt]


@istest
@with_setup(teardown=reset_config)
def startup_ramprate_default():
    """
    Create two generators with a high limit,
    one limited by Pmax, the other by rampratemax.
    Ensure that the second gen. is started up to its limit to minimize shed.
    """

    pmax = 100
    Pdt1 = 210
    generators = [
        make_cheap_gen(pmax=pmax),
        make_expensive_gen(rampratemax=100, rampratemin=-200, pmin=95, pmax=200),
    ]
    initial = [{"power": 90}, {"status": 0}]
    Pdt = [90, Pdt1, 90]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # the expensive unit turns on
    assert_series_equal(
        pd.Series([0.0, 1.0, 0.0], index=times.times),
        generators[1].values("status", times.times),
    )

    # the minimum load was shed
    assert_series_equal(
        pd.Series([0.0, 100.0, 0.0], index=times.times),
        generators[1].values("power", times.times),
    )



from minpower import solve, config
import sys


def test(resultsfile="out.txt", logfile="out.log"):
    sys.stdout = open(resultsfile, "w+")
    solve.solve_problem(
        datadir="/home/adam/ercot-data/ercot-2010-test-system/",
        shell=True,
        problemfile=False,
        visualization=False,
        csv=False,
        solver="cplex",
        num_breakpoints=10,
        hours_commitment=24,
        hours_commitment_overlap=12,
        get_duals=False,
        logging_file=logfile,
        solution_file=False,
    )


if __name__ == "__main__":
    config.updown_formulation = sys.argv[1]
    filename = config.updown_formulation.replace(" ", "-")
    test(resultsfile=filename + ".txt", logfile=filename + ".log")



import pyomo.environ as pyo
from pyomo.opt import SolverFactory
from .test_utils import istest


@istest
def test_pyomo_duals():
    # create a model
    model = pyo.ConcreteModel()
    # Create a 'dual' suffix component on the instance
    # so the solver plugin will know which suffixes to collect
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    model.x = pyo.Var([1, 2], domain=pyo.NonNegativeReals)
    model.OBJ = pyo.Objective(expr=2 * model.x[1] + 3 * model.x[2])
    model.Constraint1 = pyo.Constraint(expr=3 * model.x[1] + 4 * model.x[2] >= 1)

    opt = SolverFactory("glpk")
    opt.solve(model)

    # make sure the dual is non-zero (constraint is binding)
    assert model.dual[model.Constraint1] != 0


@istest
def test_duals_mip():
    model = pyo.ConcreteModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

    model.x = pyo.Var([1, 2], domain=pyo.Boolean)
    model.y = pyo.Var([1, 2], domain=pyo.NonNegativeReals)
    # this problem is basically just two generators:
    # gen1 is cheap, but limited output
    # gen2 is more expensive
    model.OBJ = pyo.Objective(expr=2 * model.y[1] + 3 * model.y[2])
    model.Constraint1 = pyo.Constraint(expr=model.y[1] <= 50 * model.x[1])
    model.Constraint2 = pyo.Constraint(expr=model.y[2] <= 80 * model.x[2])
    model.Constraint3 = pyo.Constraint(expr=model.y[1] + model.y[2] == 100)

    opt = SolverFactory("glpk")
    opt.solve(model)

    # make sure the solution is what we'd expect (both gen need to run)
    assert model.x[1].value == True
    assert model.x[2].value == True

    # fix binary variables
    model.x[1].fixed = True
    model.x[2].fixed = True
    opt.solve(model)
    # make sure the dual is non-zero (constraint is binding)
    assert model.dual[model.Constraint3] != 0



import logging
import nose
import numpy as np
import pandas as pd
from minpower.tests.test_utils import istest
from minpower.tests.test_integration import (
    run_case,
    assert_series_equal,
    assert_frame_equal,
)

logging.basicConfig(level=logging.ERROR)


raise nose.SkipTest("TODO: stochastic UC currently not working with new pyomo version")

mipgap = 0.0001


@istest
def run_basic_checks():
    """
    solutions should all make sense (cost=0 and power=0 if status=0)
    """
    run_case("stochastic_short_case", deterministic_solve=True)


@istest
def mock_stochastic():
    """
    the solution to the stochastic problem with only two scenarios,
    both with values equal to the deterministic forecast and prob=0.5,
    should have a total cost equal (within solver tolerance) to
    the deterministic forecat solution total cost
    """
    slnS = run_case("stochastic_mock_case", scenarios_directory="scenarios/")
    slnD = run_case("stochastic_mock_case", deterministic_solve=True)

    costS = slnS.observed_cost.sum().sum()
    costD = slnD.observed_cost.sum().sum()

    diffpercent = np.abs((costS - costD) / costD)
    assert diffpercent < mipgap


@istest
def standalone():
    try:
        import tables
    except ImportError:
        raise nose.SkipTest("standalone mode requires pytables")

    slnA = run_case("stochastic_short_case", deterministic_solve=True)
    slnB = run_case("stochastic_short_case", deterministic_solve=True, standalone=True)

    assert slnA.observed_cost.sum().sum() == slnB.observed_cost.sum().sum()


@istest
def expected_cost_case():
    """
    ensure that a under-forecast wind case has
    an expected cost > observed cost
    """
    # this is a short and simple case with 4hrs, but make it into two UC days
    hrs = dict(hours_commitment=2, hours_overlap=0)
    slnD = run_case("expected_observed_cost", deterministic_solve=True, **hrs)

    assert slnD.expected_cost.sum().sum() > slnD.observed_cost.sum().sum()


@istest
def designed_diff_case():
    """
    ensure that a simple case designed to produce a more expensive
    deterministic forecast solution has a cheaper perfect forecast
    """
    # this is a short and simple case with 4hrs, but make it into two UC days
    hrs = dict(hours_commitment=2, hours_overlap=0)
    slnP = run_case("deterministic_perfect_difference", perfect_solve=True, **hrs)
    slnD = run_case("deterministic_perfect_difference", deterministic_solve=True, **hrs)

    assert (
        slnP.generators_power.sum(axis=1) - slnD.generators_power.sum(axis=1) == 0
    ).all()
    assert slnP.observed_cost.sum().sum() < slnD.observed_cost.sum().sum()


@istest
def stochastic_load_shedding():
    """
    Create a sheddable wind generator and one scenario that requires load shedding.
    Ensure that wind is shed in one scenario but not the other.
    """

    sln = run_case("stochastic_shedding_case")

    epower = sln.generators_power_scenarios
    load_sched = sln.loads[0].schedule

    assert (epower["s0"].sum(axis=1) < load_sched).all()
    assert (epower["s1"].sum(axis=1) == load_sched).all()


@istest
def stochastic_gen_shedding():
    """
    Create a sheddable wind generator and one scenario that requires wind shedding.
    Ensure that wind is shed in one scenario but not the other.
    """

    sln = run_case("stochastic_shedding_wind_case")

    epower = sln.generators_power_scenarios
    load_sched = sln.loads[0].schedule
    load_sched.index = epower.major_axis
    # check that generation meets the load for all scenarios
    assert epower.sum(axis=2).eq(load_sched, axis=0).all().all()

    # get wind power scenarios
    scenario_values = sln.generators[0].scenario_values.drop("probability", axis=2)
    scenario_values = (
        scenario_values[scenario_values.items[0]]
        .T.loc[: len(epower.major_axis) - 1]
        .set_index(epower.major_axis)
        .rename(columns=dict(list(zip(list(range(len(sln.scenarios))), sln.scenarios))))
    )
    # calculate wind shed per scenario
    wind_shed = scenario_values - epower.minor_xs("g0")

    assert_series_equal(pd.Series(dict(s0=False, s1=True)), (wind_shed > 0).all())


# @istest
# FIXME - disabled test for now


def cvar_objective():
    """
    Set up an either-or two gen case: g1 cheap, g2 expensive,
    with two wind senarios:
        s0 - high prob., use cheap gen
        s1 - very low prob. (below conf. limit), expensive gen needed

    CVaR should heavily weight the unlikely but more expensive scenario.
    EV must consider it, but weights it less.
    The result should be a more expensive CVaR objective.
    """
    slnEV = run_case("stochastic_cvar_case")
    slnCV = run_case(
        "stochastic_cvar_case",
        cvar_weight=1.0,  # very risk averse
        cvar_confidence_level=0.9999,
    )  # very long tail

    assert slnCV.objective > slnEV.objective
    #    try:
    assert_frame_equal(slnCV.expected_totalcost, slnEV.expected_totalcost)


#    except:
#        debug_frame_unequal(slnCV.expected_totalcost, slnEV.expected_totalcost)



"""Test the constraint behavior of the generators"""

import logging

logging.basicConfig(level=logging.CRITICAL, format="%(levelname)s: %(message)s")

import pandas as pd
import numpy as np

from pandas.testing import assert_series_equal
from minpower.generators import Generator_nonControllable, Generator_Stochastic
from minpower.optimization import value
from minpower.schedule import TimeIndex
from .test_utils import *


@istest
def power_maximum():
    """
    Create two generators, one cheap and one expensive.
    Give the cheap one a maximum power limit lower than the load.
    Ensure that the cheaper unit is at its limit.
    """
    pmax = 100
    Pd = 221
    generators = [make_cheap_gen(pmax=pmax), make_expensive_gen()]
    _, times = solve_problem(generators, **make_loads_times(Pd))
    assert value(generators[0].power(times[0])) == pmax


@istest
def power_minimum():
    """
    Create two generators, one cheap and one expensive, status on for both.
    Give the expensive one a min power limit.
    Ensure that the exp. unit is at its limit.
    """
    pmin = 10
    Pd = 221
    generators = [make_cheap_gen(), make_expensive_gen(pmin=pmin)]
    _, times = solve_problem(generators, **make_loads_times(Pd))
    assert value(generators[1].power(times[0])) == pmin


@istest
def ramp_up():
    """
    Create 2 gen, cheaper one with ramp up limit.
    Create load that ramps up faster than limit, at t1.
    Ensure that cheaper unit is at its limit.
    """
    ramp_limit = 80
    generators = [make_cheap_gen(rampratemax=ramp_limit), make_expensive_gen()]
    initial = [{"power": 250}, {"power": 250}]
    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[250, 350])
    )
    print((generators[1]._parent_problem()))
    assert (
        value(generators[0].power(times[1])) - value(generators[0].power(times[0]))
        == ramp_limit
    )


@istest
def ramp_down():
    """
    Create 2 gen, expensive one with ramp down limit.
    Create load that ramps down faster than limit at t1.
    Ensure that exp. unit is at its limit.
    """
    ramp_limit = -80
    generators = [make_cheap_gen(), make_expensive_gen(rampratemin=ramp_limit)]
    initial = [{"power": 250}, {"power": 250}]
    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[550, 450])
    )
    ramp_rate = value(generators[1].power(times[1])) - value(
        generators[1].power(times[0])
    )
    assertAlmostEqual(ramp_rate, ramp_limit)


@istest
def ramp_up_initial():
    """
    Create 2 gen, cheaper one with ramp up limit.
    Create load that ramps up faster than limit, at t0.
    Ensure that cheaper unit is at its limit.
    """
    ramp_limit = 80
    generators = [make_cheap_gen(rampratemax=ramp_limit), make_expensive_gen()]
    initial = [{"power": 250}, {"power": 250}]

    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[350, 350])
    )
    ramp_rate = value(generators[0].power(times[0])) - value(
        generators[0].initial_power
    )
    assert ramp_rate == ramp_limit


@istest
def ramp_down_initial():
    """
    Create 2 gen, expensive one with ramp down limit.
    Create load that ramps down faster than limit, at t0.
    Ensure that exp. unit is at its limit.
    """
    ramp_limit = -80
    generators = [make_cheap_gen(), make_expensive_gen(rampratemin=ramp_limit)]
    initial = [{"power": 250}, {"power": 250}]

    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[300, 300])
    )
    ramp_rate = value(generators[1].power(times[0])) - value(
        generators[1].initial_power
    )
    assert ramp_rate == ramp_limit


@istest
def cold_ramp_up():
    """
    Create three generators, one that must ramp up from cold.
    Ensure that the generator ramps at within its limit.
    """
    ramp_limit_SU = 20
    ramp_limit = 30
    generators = [
        make_cheap_gen(pmax=200),
        make_mid_gen(pmin=10, rampratemax=ramp_limit, startupramplimit=ramp_limit_SU),
        make_expensive_gen(pmin=1),
    ]
    initial = [{"power": 200}, {"status": 0}, {"status": 0}]

    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[200, 230, 280])
    )

    ramp_rate_SU = value(generators[1].power(times[0])) - generators[1].initial_power
    assert ramp_rate_SU <= ramp_limit_SU
    ramp_rate = value(generators[1].power(times[2])) - value(
        generators[1].power(times[1])
    )
    assert ramp_rate <= ramp_limit


@istest
def hot_shut_down():
    """
    Create three generators, one that must shut down from hot.
    Ensure that the generator ramps within its limit.
    """
    ramp_limit_SD = -30
    ramp_limit = -20
    generators = [
        make_cheap_gen(pmax=200),
        make_expensive_gen(
            pmin=20, rampratemin=ramp_limit, shutdownramplimit=ramp_limit_SD
        ),
        make_mid_gen(pmin=1),
    ]
    initial = [{"power": 200}, {"power": 50}, {"power": 50}]

    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[280, 150])
    )

    ramp_rate = value(generators[1].power(times[0])) - generators[1].initial_power
    assert ramp_limit <= ramp_rate

    ramp_rate_SD = value(generators[1].power(times[1])) - value(
        generators[1].power(times[0])
    )
    assert ramp_limit_SD == ramp_rate_SD
    assert value(generators[1].status(times[1])) == 0


@istest
def min_up_time():
    """
    Create two generators: cheap with max power limit, exp. with min up time and min power limits.
    Create load that increases over the cheap limit at t1 and then reduces back to constant.
    Ensure that the expensive generator is on at t1 and t2, then turns off.
    """
    generators = [make_cheap_gen(pmax=100), make_expensive_gen(minuptime=2, pmin=5)]
    initial = [dict(power=80, status=True), dict(status=False)]
    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[85, 120, 80, 80])
    )
    limgen_status = [value(generators[1].status(t)) for t in times]
    assert limgen_status == [0, 1, 1, 0] or limgen_status == [1, 1, 0, 0]


@istest
def min_down_time():
    """
    Create three generators:
    * cheap with max power limit
    * with min down time
    * expensive
    Create load that is higher than the cheap limit all but t1.
    Ensure that the mid generator is OFF at t1 and t2, then turns on again.
    Ensure that the hi generator is ON for t2 to make up the difference.
    """
    generators = [
        make_cheap_gen(pmax=100),
        make_mid_gen(mindowntime=2, pmin=40),
        make_expensive_gen(),
    ]
    initial = [
        dict(power=100, status=True),
        dict(power=40, status=True, hoursinstatus=0),
        dict(status=False),
    ]

    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[150, 10, 140, 140])
    )
    limgen_status = [value(generators[1].status(t)) for t in times]
    expensive_status_t2 = value(generators[2].status(times[2]))
    assert limgen_status == [1, 0, 0, 1] and expensive_status_t2 == 1


@istest
def start_up_cost():
    """
    Create two generators:
    * cheap with a max power
    * expensive with a start up cost.
    Create an increasing load.
    Ensure that the objective includes the startup cost.
    """
    startupcost = 9000
    generators = [
        make_cheap_gen(pmax=100),
        make_expensive_gen(pmin=10, startupcost=startupcost),
    ]
    initial = [dict(power=80, status=True), dict(status=False)]
    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[80, 120])
    )
    assert value(generators[1].cost_startup(times[1])) == startupcost


@istest
def shut_down_cost():
    """
    Create two generators, both on:
    * cheap with a max power
    * expensive with a start down cost and a min power.
    Create an decreasing load.
    Ensure that the objective includes the shutdown cost.
    """
    shutdowncost = 200
    generators = [
        make_cheap_gen(pmax=100),
        make_expensive_gen(shutdowncost=shutdowncost, pmin=20),
    ]
    initial = [dict(power=80, status=True), dict(power=20, status=True)]
    _, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=[150, 10])
    )
    assert value(generators[1].cost_shutdown(times[1])) == shutdowncost


@istest
def min_up_time_longer():
    """
    Create two generators: cheap with max power limit, exp. with min up time and min power limits.
    Create load that increases over the cheap limit at t1 and then reduces back to constant.
    Ensure that the expensive generator is on at t1 and t2, then turns off.
    """
    generators = [make_cheap_gen(pmax=100), make_expensive_gen(minuptime=8, pmin=5)]
    initial = [
        dict(power=80, status=True, hoursinstatus=1),
        dict(status=False, hoursinstatus=0),
    ]
    _, times = solve_problem(
        generators,
        gen_init=initial,
        **make_loads_times(Pdt=[85, 120, 80, 80, 70, 70, 70, 70, 80, 80])
    )
    limgen_status = [value(generators[1].status(t)) for t in times]
    # logging.critical(limgen_status)
    # logging.critical([(generators[1].startup(t),generators[1].shutdown(t)) for t in times])
    # logging.critical(problem.constraints['minuptime_g1t01'])
    assert limgen_status == [0, 1, 1, 1, 1, 1, 1, 1, 1, 0] or limgen_status == [
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
    ]


@istest
def wind_shedding():
    """
    Create a sheddable wind generator and a situation that requires shedding.
    Ensure that the minimum power is shed.
    """
    lts = make_loads_times(Pdt=[85, 110, 80, 80])
    Pwind = pd.Series([60, 100, 120, 60], index=lts["times"].strings.values)

    generators = [
        Generator_nonControllable(schedule=Pwind, sheddingallowed=True),
        make_expensive_gen(),
    ]

    _, times = solve_problem(generators, **lts)
    assert value(generators[0].power(times[2])) == 80
    assert value(generators[0].power(times[3])) == 60
    assert sum(value(generators[1].power(t)) for t in times) == 25 + 10 + 20

    # this should work for stochastic generators
    # TODO - a real stochastic test case
    #   this one just tests deterministic solve
    user_config.deterministic_solve = True
    generators = [
        Generator_Stochastic(schedule=Pwind, sheddingallowed=True),
        make_expensive_gen(),
    ]
    _, times = solve_problem(generators, **lts)
    assert value(generators[0].power(times[2])) == 80
    assert value(generators[0].power(times[3])) == 60
    assert sum(value(generators[1].power(t)) for t in times) == 25 + 10 + 20


@istest
def pmin_startup_limit():
    """
    generators that have a pmin > rampratemax should be able to startup to pmin
    """
    pmin = 185
    generators = [
        make_cheap_gen(pmin=0, pmax=100),
        make_mid_gen(pmin=pmin, pmax=200, rampratemax=10),
        make_expensive_gen(pmin=0, pmax=200),
    ]
    initial = [{"power": 100}, {"status": 0}, {"power": 85}]
    lts = make_loads_times(Pdt=[185, 210, 250, 285, 290])

    power_system, times = solve_problem(generators, gen_init=initial, **lts)

    power = generators[1].values("power")
    status = generators[1].values("status")
    assert (power[status == 1] >= pmin).all()
    assert generators[1].values("status")[times[4]] == 1


@istest
def pmin_shutdown_limit():
    """
    generators that have a pmin > -1 * rampratemin should be able to shutdown
    """
    pmin = 185
    generators = [
        make_cheap_gen(pmin=0, pmax=200),
        make_mid_gen(pmin=pmin, pmax=200, rampratemin=-10),
    ]
    initial = [{"power": 100}, {"power": 200}]
    lts = make_loads_times(Pdt=[300, 250, 180])

    power_system, times = solve_problem(generators, gen_init=initial, **lts)

    status = generators[1].values("status")
    assert status[times[2]] == 0


@istest
def startupcost_can_shutdown():
    """
    Generators with a startup cost must be able to shut down.
    """
    pmax = 100
    startupcost = 100
    generators = [
        make_cheap_gen(pmax=pmax),
        make_expensive_gen(startupcost=startupcost, pmin=20),
    ]
    initial = [{"power": 90}, {"status": 0}]
    Pdt = [90, 120, 90]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # the expensive unit turns on and back off
    assert_series_equal(
        pd.Series([0.0, 1.0, 0.0], index=times.strings), generators[1].values("status")
    )

    assert_series_equal(
        pd.Series([0.0, startupcost, 0.0], index=times.strings),
        generators[1].values("startupcost"),
    )


@istest
def setting_initial_conditions():
    """
    make sure that power and status get properly set (even if numpy values)
    """
    gen = make_cheap_gen()

    # numpy values should be converted to floats
    gen.set_initial_condition(power=np.float64(5.0), status=1)
    assert gen.initial_power == 5.0
    assert type(gen.initial_power) == float

    # status must get conveted to int
    gen.set_initial_condition(status=True)
    assert gen.initial_status == 1
    assert type(gen.initial_status) == int

    # if status = 0, power must be zero
    gen.set_initial_condition(power=5.0, status=0)
    assert gen.initial_power == 0


@istest
def final_conditions():
    """make sure that Generator.gethrsinstatus calculates correctly"""
    gen = make_cheap_gen()
    gen.set_initial_condition(status=1, hoursinstatus=10)

    times = TimeIndex(
        pd.date_range("2010-01-01 00:00:00", "2010-01-01 5:00:00", freq="H")
    )

    # status varies
    status = pd.Series([1, 0, 1, 0, 0, 0], index=times)
    assert gen.gethrsinstatus(times, status) == 3

    # status varies, equal to initial status
    status = pd.Series([0, 0, 0, 1, 1, 1], index=times)
    assert gen.gethrsinstatus(times, status) == 3

    # status is all the same and equal to the initial status
    status = pd.Series(1, index=times)
    assert gen.gethrsinstatus(times, status) == 16

    # status is all the same and not equal to the initial status
    status = pd.Series(0, index=times)
    assert gen.gethrsinstatus(times, status) == 6


@istest
def initial_min_up_time():
    """ensure the generator meets its minuptime limit at t0"""
    pmax = 100
    generators = [make_cheap_gen(pmax=pmax), make_expensive_gen(minuptime=2, pmin=10)]
    initial = [
        {"power": 90},
        {"power": 10, "hoursinstatus": 1},
    ]
    Pdt = [90, 90]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # the expensive unit can't turn off in the first hour
    # but it can turn off in the second hour
    assert_series_equal(
        pd.Series([1, 0], index=times.strings),
        generators[1].values("status").astype(int),
    )


@istest
def initial_min_down_time():
    """ensure the generator meets its mindowntime limit at t0"""
    pmax = 100
    generators = [
        make_cheap_gen(pmax=pmax),
        make_mid_gen(mindowntime=2),
        make_expensive_gen(pmin=1),
    ]
    initial = [
        {"power": 90},
        {"status": 0, "hoursinstatus": 1},
        {"power": 10},
    ]
    Pdt = [110, 120]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # the expensive unit can't turn on in the first hour
    # but it can turn on in the second hour
    assert_series_equal(
        pd.Series([0, 1], index=times.strings),
        generators[1].values("status").astype(int),
    )

    # the first hr difference has to be picked up by the more expensive unit
    assert_series_equal(
        pd.Series([1, 0], index=times.strings),
        generators[2].values("status").astype(int),
    )


@istest
def final_min_up_time():
    """ensure the generator meets its minuptime limit at t_final"""
    pmax = 100
    generators = [make_cheap_gen(pmax=pmax), make_expensive_gen(minuptime=6, pmin=10)]
    initial = [
        {"power": 90},
        {"power": 10, "hoursinstatus": 3},
    ]
    Pdt = [90, 90, 90]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # the expensive unit can't turn off
    assert_series_equal(
        pd.Series([1, 1, 1], index=times.strings),
        generators[1].values("status").astype(int),
    )

    # now, set the limit to 5hrs
    generators[1].minuptime = 5
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # and the expensive unit can turn off in the last hour
    assert_series_equal(
        pd.Series([1, 1, 0], index=times.strings),
        generators[1].values("status").astype(int),
    )


@istest
def final_min_down_time():
    """ensure the generator meets its minuptime limit at t_final"""
    pmax = 100
    generators = [
        make_cheap_gen(pmax=pmax),
        make_mid_gen(mindowntime=6),
        make_expensive_gen(pmin=1),
    ]
    initial = [
        {"power": 90},
        {"status": 0, "hoursinstatus": 3},
        {"power": 10},
    ]
    Pdt = [110, 110, 110]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # the mid unit can't turn on
    assert_series_equal(
        pd.Series([0, 0, 0], index=times.strings),
        generators[1].values("status").astype(int),
    )

    # now, set the limit to 5hrs
    generators[1].mindowntime = 5
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt)
    )

    # and the mid unit can turn on in the last hour
    assert_series_equal(
        pd.Series([0, 0, 1], index=times.strings),
        generators[1].values("status").astype(int),
    )



from minpower.tests.test_utils import user_config, istest, with_setup, reset_config
from minpower.get_data import parsedir
import os


basedir = os.path.split(__file__)[0]


@istest
@with_setup(teardown=reset_config)
def check_pmin():
    user_config.directory = os.path.join(basedir, "uc")
    generators, loads, _, times, _, data = parsedir()

    assert data["generators"].pmin.tolist() == [gen.pmin for gen in generators]



"""Test the constraint behavior of an OPF"""
from minpower import powersystems
from minpower.optimization import value
from minpower.commonscripts import Series
from minpower.config import user_config

from .test_utils import (
    istest,
    get_duals,
    with_setup,
    reset_config,
    make_cheap_gen,
    make_mid_gen,
    make_expensive_gen,
    singletime,
    make_loads_times,
    solve_problem,
)


def test_config():
    assert user_config.duals == False


@istest
@with_setup(get_duals, reset_config)
def line_limit_high():
    """
    Create a two bus system
        -cheap gen on busA
        -expensive gen and load on busB,
        -line limit lower than load
    Ensure that:
        - the line is at its limit
        - the congestion price is equal to the diff. in LMPs
    """
    pmax = 100
    generators = [make_cheap_gen(bus="A"), make_expensive_gen(bus="B")]

    lines = [powersystems.Line(pmax=pmax, frombus="A", tobus="B")]
    power_system, times = solve_problem(
        generators,
        do_reset_config=False,
        lines=lines,
        **make_loads_times(Pd=225, bus="B")
    )
    Pline = value(lines[0].power(times[0]))
    lmps = [b.price(times[0]) for b in power_system.buses]
    congestion_price = lines[0].price(times[0])
    assert Pline == pmax and congestion_price == (lmps[1] - lmps[0])


@istest
@with_setup(get_duals, reset_config)
def line_limit_low():
    """
    Create a two bus system
        -cheap gen on busA
        -expensive gen and load on busB,
        -line (from B to A) limit lower than load
    Ensure that:
        - the line is at its limit
        - the congestion price is equal to the diff. in LMPs
    """
    pmin = -100
    generators = [make_cheap_gen(bus="A"), make_expensive_gen(bus="B")]
    lines = [powersystems.Line(pmin=pmin, frombus="B", tobus="A")]
    power_system, times = solve_problem(
        generators,
        do_reset_config=False,
        lines=lines,
        **make_loads_times(Pd=225, bus="B")
    )
    Pline = value(lines[0].power(times[0]))
    lmps = [b.price(times[0]) for b in power_system.buses]
    congestion_price = lines[0].price(times[0])
    assert Pline == pmin and congestion_price == -1 * (lmps[1] - lmps[0])


@istest
@with_setup(get_duals, reset_config)
def three_buses():
    """
    Create a three bus system
        -cheap gen and load on busA
        -mid gen and load on busB
        -expensive gen and load on busC
        -three lines connecting all three buses (triangle)
        -create congestion by setting a line limit
    Ensure that:
        - the total load is met
        - the lmps are different
    """
    pmax = 50
    Pd = [105, 225, 302]
    generators = [
        make_cheap_gen(bus="A"),
        make_mid_gen(bus="B"),
        make_expensive_gen(bus="C"),
    ]
    loads = [
        powersystems.Load(schedule=Series(Pd[0], singletime), bus="A"),
        powersystems.Load(schedule=Series(Pd[1], singletime), bus="B"),
        powersystems.Load(schedule=Series(Pd[2], singletime), bus="C"),
    ]
    lines = [
        powersystems.Line(frombus="A", tobus="B"),
        powersystems.Line(frombus="A", tobus="C", pmax=pmax),
        powersystems.Line(frombus="B", tobus="C", pmax=pmax),
    ]
    power_system, times = solve_problem(
        generators, do_reset_config=False, times=singletime, loads=loads, lines=lines
    )
    total_load = value(sum(b.Pload(times[0]) for b in power_system.buses))
    assert total_load == sum(Pd)
    num_lmps = len(set(b.price(times[0]) for b in power_system.buses))
    assert num_lmps > 1


def test_config_cleared():
    assert user_config.duals == False



"""
Stochastic scenario models for schedules.
"""
from pyomo.environ import AbstractModel, Set, Param, Boolean, Var
from .config import user_config
import gc
import logging


def construct_simple_scenario_tree(power_system, times, time_stage=None):
    """Construct a simple scenario tree instance"""

    gen = power_system.get_generator_with_scenarios()
    probabilities = gen._get_scenario_probabilities(times).values.tolist()

    tree = new_scenario_tree_model()

    prob_set = list(range(len(probabilities)))
    if time_stage is None:
        scenario_names = ["s{n}".format(n=n) for n in prob_set]
        node_names = ["n{n}".format(n=n) for n in prob_set]
    else:
        scenario_names = ["s{n}t{t}".format(n=n, t=time_stage) for n in prob_set]
        node_names = ["n{n}t{t}".format(n=n, t=time_stage) for n in prob_set]

    tree.Stages.add("first stage", "second stage")
    tree.Nodes.add("root")
    for nm in node_names:
        tree.Nodes.add(nm)
    tree.NodeStage["root"] = "first stage"
    for nm in node_names:
        tree.NodeStage[nm] = "second stage"
    tree.Children["root"] = node_names

    tree.ConditionalProbability["root"] = 1.0
    for i, pr in enumerate(probabilities):
        tree.ConditionalProbability[node_names[i]] = pr

    for nm in scenario_names:
        tree.Scenarios.add(nm)
    for i, nm in enumerate(scenario_names):
        tree.ScenarioLeafNode[nm] = node_names[i]
    # tree.Scenarios.pprint()

    power_system._scenario_tree_instance = tree
    logging.debug("constructed tree for stage {s}".format(s=time_stage))
    # note that stage variables still need definition
    # as do the values of the per scenario variables
    return


def define_stage_variables(power_system, times):
    # scenario_tree.Stages.pprint()
    # scenario_tree.StageVariables.pprint()

    # create sets of variable names (not actual variables) for each stage
    variables_first_stage = Set()
    variables_second_stage = Set()

    for gen in power_system.get_generators_without_scenarios():
        # for each non stochastic generator
        if gen.is_controllable:
            variables_first_stage.add(
                str(gen.get_variable("status", indexed=True, time=None)) + "[*]"
            )
            variables_second_stage.add(
                str(gen.get_variable("power", indexed=True, time=None)) + "[*]"
            )
        # note - appending '[*]' to the indicies is required to get
        # pysp to assign all the variables in the array to a shape

    if power_system.shedding_mode:
        for gen in [
            gen
            for gen in power_system.get_generators_noncontrollable()
            if gen.shedding_mode
        ]:
            variables_second_stage.add(
                str(gen.get_variable("power_used", time=None, indexed=True)) + "[*]"
            )
        for load in power_system.loads():
            variables_second_stage.add(
                str(load.get_variable("power", time=None, indexed=True)) + "[*]"
            )

    # variables_first_stage.pprint()
    scenario_tree = power_system._scenario_tree_instance
    scenario_tree.StageDerivedVariables = []
    scenario_tree.StageVariables["first stage"] = variables_first_stage
    scenario_tree.StageVariables["second stage"] = variables_second_stage

    scenario_tree.StageCostVariable["first stage"] = str(
        power_system.cost_first_stage()
    )
    scenario_tree.StageCostVariable["second stage"] = str(
        power_system.cost_second_stage()
    )


def create_problem_with_scenarios(power_system, times):
    from pysp.scenariotree import ScenarioTree
    from pysp.ef import create_ef_instance

    logging.debug("constructing scenario tree")

    scenario_tree = ScenarioTree(
        scenarioinstance=power_system._model,
        scenariotreeinstance=power_system._scenario_tree_instance,
    )

    if not scenario_tree.validate():
        for s, scenario in enumerate(scenario_tree._scenarios):
            print((s, scenario))
        raise ValueError("not a valid scenario tree")

    gen = power_system.get_generator_with_scenarios()

    scenario_instances = {}

    # construct scenario instances
    logging.debug("constructing scenario instances")
    gc.disable()
    for s, scenario in enumerate(scenario_tree._scenarios):
        scenario_instance = power_system._model.clone()

        power = getattr(scenario_instance, "power_{}".format(str(gen)))
        # set the values of the parameter for this scenario
        logging.debug("setting scenario values for s%i" % s)
        scenario_vals = gen._get_scenario_values(times, s=s)
        for t, time in enumerate(times):
            power[time] = scenario_vals[t]

        # power.pprint()
        scenario_instance.preprocess()
        scenario_instances[scenario._name] = scenario_instance

    gc.enable()
    scenario_tree.defineVariableIndexSets(power_system._model)

    cvar_params = {}
    if user_config.cvar_weight > 0:
        cvar_params = dict(
            generate_weighted_cvar=True,
            cvar_weight=user_config.cvar_weight,
            risk_alpha=user_config.cvar_confidence_level,
        )

    full_problem_instance = create_ef_instance(
        scenario_tree, scenario_instances, **cvar_params
    )

    # full_problem_instance.pprint()

    # relax the non-anticipatory constraints on the generator status variables
    # beyond the UC time horizon
    if len(times.post_horizon()) > 0:
        logging.debug(
            "removing non-anticipatory constraints for post-commitment horizon"
        )
    for time in times.post_horizon():
        for scenario in list(scenario_instances.keys()):
            for gen in power_system.generators():
                if not gen.is_controllable:
                    continue
                u = gen.status().name
                delattr(
                    full_problem_instance,
                    "{s}_{u}_{t}".format(s=scenario, u=u, t=str(time)),
                )

    power_system.stochastic_formulation = True
    power_system._stochastic_instance = full_problem_instance
    power_system._scenario_tree = scenario_tree
    power_system._scenario_instances = scenario_instances
    return


def get_scenario_based_costs(scenario_tree, scenario_instances):
    # scenario_tree.pprintCosts(scenario_instances)
    costs = dict()
    for node in list(scenario_tree._tree_node_map.values()):
        scenarios = node._scenarios
        if len(scenarios) == 1:
            costs[node._scenarios[0]._name] = node.computeExpectedNodeCost(
                scenario_instances
            )
        else:
            continue
    return costs


def update_variables(power_system, times):
    """Convert all variables into dictionaries of their solved values, keyed by scenario"""
    variable_names = list(
        power_system._model.component_objects(Var, active=True).keys()
    )
    values = dict([(nm, {}) for nm in variable_names])
    for s, scenario in list(power_system._scenario_instances.items()):
        for var_name, var in list(scenario.component_objects(Var, active=True).items()):
            if var.is_indexed():
                values[var_name][s] = dict(
                    [(idx, var_val.value) for idx, var_val in list(var.items())]
                )
            else:
                values[var_name][s] = var.value
    power_system._per_scenario_values = values


def new_scenario_tree_model():
    # https://software.sandia.gov/trac/coopr/browser/coopr.pysp/trunk/coopr/pysp/util/scenariomodels.py
    scenario_tree_model = AbstractModel()

    # all set/parameter values are strings, representing the names of various
    # entities/variables.

    scenario_tree_model.Stages = Set(ordered=True)
    scenario_tree_model.Nodes = Set()

    scenario_tree_model.NodeStage = Param(
        scenario_tree_model.Nodes, within=scenario_tree_model.Stages
    )
    scenario_tree_model.Children = Set(
        scenario_tree_model.Nodes, within=scenario_tree_model.Nodes, ordered=True
    )
    scenario_tree_model.ConditionalProbability = Param(scenario_tree_model.Nodes)

    scenario_tree_model.Scenarios = Set(ordered=True)
    scenario_tree_model.ScenarioLeafNode = Param(
        scenario_tree_model.Scenarios, within=scenario_tree_model.Nodes
    )

    scenario_tree_model.StageVariables = Set(scenario_tree_model.Stages)
    scenario_tree_model.StageCostVariable = Param(scenario_tree_model.Stages)

    # scenario data can be populated in one of two ways. the first is "scenario-based",
    # in which a single .dat file contains all of the data for each scenario. the .dat
    # file prefix must correspond to the scenario name. the second is "node-based",
    # in which a single .dat file contains only the data for each node in the scenario
    # tree. the node-based method is more compact, but the scenario-based method is
    # often more natural when parameter data is generated via simulation. the default
    # is scenario-based.
    scenario_tree_model.ScenarioBasedData = Param(
        within=Boolean, default=True, mutable=True
    )

    # do we bundle, and if so, how?
    scenario_tree_model.Bundling = Param(within=Boolean, default=False, mutable=True)
    scenario_tree_model.Bundles = Set()  # bundle names
    scenario_tree_model.BundleScenarios = Set(scenario_tree_model.Bundles)

    return scenario_tree_model



from .test_utils import *
from minpower.results import make_solution


@istest
@with_setup(teardown=reset_config)
def check_power_status():
    """
    Ensure that the solution generator_status and generator_power attributes
    match with the optimization values returned.
    """
    generators = [make_cheap_gen(pmax=100), make_mid_gen(pmax=20), make_expensive_gen()]

    user_config.duals = True
    power_system, times = solve_problem(
        generators, **make_loads_times(Pdt=[80, 110, 130])
    )

    sln = make_solution(power_system, times)
    t0 = sln.generators_power.index[0]

    assert sln.generators_power.loc[t0, "g0"] == value(generators[0].power(times[0]))

    assert sln.generators_status.loc[t0, "g1"] == value(generators[1].status(times[0]))



"""Test for memory leaks"""
from minpower import powersystems, solve
from .test_utils import solve_problem, make_loads_times
import objgraph
import inspect


def show_memory_backrefs(name):
    objgraph.show_backrefs(
        objgraph.by_type(name), filename="backrefs-{}.png".format(name)
    )


def show_memory_refs(name):
    try:
        obj = objgraph.by_type(name)[0]
    except IndexError:
        print(("no object of type", name))
        return
    objgraph.show_chain(
        objgraph.find_backref_chain(obj, inspect.ismodule),
        filename="chain-{}.png".format(name),
    )


def show_memory_growth():
    objgraph.show_growth()


def get_counts(prefix=""):
    test_types = [
        "Var",
        "Piecewise",
        "ScenarioTree",
        "ScenarioTreeNode",
        "Scenario",
        "_SetContainer",
        "_ConstraintArray",
    ]

    # use objgraph to check if these pyomo objects still exist
    objects = {}
    test_counts = {}
    for name in test_types:
        objects[name] = objgraph.by_type(name)
        test_counts[name] = len(objects[name])

    if True:
        for name in test_types:
            if test_counts[name] == 0:
                continue
            else:
                obj = objects[name][0]
            fname = prefix + "-" + "objgraph-{}-".format(name)
            objgraph.show_refs([obj], filename=fname + "refs.png")  # too_many=50,
            objgraph.show_backrefs([obj], filename=fname + "backref.png")
            objgraph.show_chain(
                objgraph.find_backref_chain(obj, inspect.ismodule),
                filename=fname + "chain.png",
            )
    return test_counts


def leak_on_reset():
    # create a problem
    loads_times = make_loads_times(Pdt=[20, 30, 40, 50])
    generators = [
        powersystems.Generator(name="expensive", costcurveequation="30P+0.01P^2"),
        powersystems.Generator(name="cheap", costcurveequation="20P+0.005P^2"),
    ]

    # solve it
    power_system, times = solve_problem(generators, **loads_times)
    # reset the model - no pyomo components should persist
    power_system.reset_model()

    test_counts = get_counts("uc")
    assert sum(test_counts.values()) == 0


def leak_on_stochastic_reset():

    solve.solve_problem(
        datadir="./uc-stochastic",
        shell=False,
        csv=False,
        hours_commitment=24,
        hours_commitment_overlap=12,
        get_duals=False,
        Nscenarios=None,
    )

    test_counts = get_counts("uc-stochastic")
    assert sum(test_counts.values()) == 0


if __name__ == "__main__":
    leak_on_reset()



"""Test the constraint behavior of the bids"""
from minpower.generators import Generator
from minpower.optimization import value
from minpower.bidding import parse_polynomial
from .test_utils import *


@istest
def parser():
    assert parse_polynomial("7x^2 + 6x - 5") == [-5, 6, 7]
    # with explicit * symbol for multiplication
    assert parse_polynomial("7*x^2 + 6 * x - 5") == [-5, 6, 7]
    # random order
    assert parse_polynomial("6*P - 5 + 7*P^2") == [-5, 6, 7]
    # make sure we can skip the linear term
    assert parse_polynomial("6P") == [0, 6]


@istest
def linear():
    """
    Create a linear bid curve for one generators.
    Ensure that correct cost is valued for the load.
    """
    a = 5
    b = 30
    Pd = 221
    generators = [Generator(costcurveequation="{}+{}P".format(a, b))]
    _, times = solve_problem(generators, **make_loads_times(Pd))
    cost = generators[0].bids.output(times[0], evaluate=True)
    assert cost == a + b * Pd


@istest
@with_setup(teardown=reset_config)
def cubic_convex():
    """
    Create a cubic, convex bid curve for one generators.
    Ensure that linearized cost is within +5% of the true cost
    """
    Pd = 221
    a = 5
    b = 30
    c = 0.2
    d = 0.1
    user_config.breakpoints = 10

    generators = [Generator(costcurveequation="{}+{}P+{}P^2+{}P^3".format(a, b, c, d))]
    _, times = solve_problem(
        generators, **make_loads_times(Pd)
    )  # ,problem_filename='bidproblem.lp')
    cost = value(generators[0].bids.output(times[0], evaluate=True))
    actual_cost = a + b * Pd + c * Pd ** 2 + d * Pd ** 3
    assert actual_cost <= cost and cost <= 1.05 * actual_cost


@istest
def cubic_non_convex():
    """
    Create a cubic, but non-convex (negative cubic term) bid curve for one generators.
    Ensure that linearized cost is within +5% of the true cost
    """
    Pd = 221
    a = 5
    b = 30
    c = 0.2
    d = 0.0001
    generators = [
        Generator(costcurveequation="{}+{}P+{}P^2 - {}P^3".format(a, b, c, d))
    ]
    power_system, times = solve_problem(generators, **make_loads_times(Pd))

    cost = generators[0].bids.output(times[0], evaluate=True)
    actual_cost = a + b * Pd + c * Pd ** 2 + -1 * d * Pd ** 3
    assert actual_cost <= cost <= 1.05 * actual_cost


@istest
def coverage():
    """
    Make sure that the bid range covers the whole pmin-pmax range
    """
    Pd = 221
    a = 5
    b = 30
    c = 0.2
    d = 0.1
    pmin = 10
    pmax = 559
    generators = [
        Generator(
            costcurveequation="{}+{}P+{}P^2+{}P^3".format(a, b, c, d),
            pmin=pmin,
            pmax=pmax,
        )
    ]
    _, times = solve_problem(generators, **make_loads_times(Pd))
    bid_points = generators[0].bids.discrete_input_points
    assert pmin == bid_points[0] and pmax == bid_points[-1]


@istest
def fixed_costs_when_off():
    """
    ensure that generator with fixed cost
    only charges fixed cost when on
    """
    a = 5
    b = 30
    c = 0.2

    generators = [
        make_cheap_gen(pmax=80),
        make_mid_gen(pmax=20),
        make_expensive_gen(
            costcurveequation="{}+{}P+{}P^2".format(a, b, c), mindowntime=1, pmax=50
        ),
    ]

    Pdt = [80, 90, 130]

    power_system, times = solve_problem(
        generators,
        gen_init=[{"power": 80}, {"power": 0}, {"status": False, "hoursinstatus": 0}],
        **make_loads_times(Pdt=Pdt)
    )

    assert generators[2].cost(times[0], evaluate=True) == 0


if __name__ == "__main__":
    nose.runmodule(argv=[__file__, "-vvs", "-x", "--pdb", "--pdb-failure"], exit=False)



import sys
import argparse
import subprocess
import os
from pprint import pprint
from minpower.config import parse_command_line_config, scheduler_config, user_config


def main():
    """
    a wrapper to make the minpower script work with cluster schedulers
    or with an ssh call that you start on your laptop and then close the
    connection.
    """

    default_minpower_config = dict(user_config).copy()
    parser = argparse.ArgumentParser(
        description="Minpower scheduler command line interface"
    )

    parser.add_argument(
        "--scheduler_mode",
        default=scheduler_config.scheduler_mode,
        choices=["qsub", "nohup", "pass"],
        help="""Mode of scheduler operation:
        qsub: use the qsub cluster scheduler
        nohup: use nohup, call as subprocess, and redirect stdin, stderr
        pass: just call minpower as a child process (for debugging)
        """,
    )
    parser.add_argument(
        "--verbose", action="store_true", default=scheduler_config.verbose
    )
    parser.add_argument(
        "--dry_run", action="store_true", help="just show, don't do", default=False
    )

    qsub_opts = parser.add_argument_group(
        "qsub options", description="the following options only apply if using qsub"
    )
    qsub_opts.add_argument(
        "--email",
        default=scheduler_config.email,
        help="email on scheduler job completion",
    )
    qsub_opts.add_argument(
        "--memory",
        type=int,
        default=scheduler_config.memory,
        help="gigabytes of memory to limit job to",
    )
    qsub_opts.add_argument(
        "--hours_limit",
        type=int,
        default=scheduler_config.hours_limit,
        help="hours of runtime to limit job to",
    )

    args, minpower_args_raw = parser.parse_known_args()

    minpower_parser = argparse.ArgumentParser("minpower")
    minpower_args = parse_command_line_config(
        minpower_parser, preparsed_args=minpower_args_raw
    )

    # scheduler config can be loaded from the case directory
    parser.set_defaults(**dict(scheduler_config))
    args, __ = parser.parse_known_args()

    # subprocess style
    minpower_args["standalone"] = True
    minpower_args["pid"] = minpower_args["pid"] if minpower_args["pid"] else os.getpid()

    if args.scheduler_mode == "qsub":
        # qsub makes all of its script calls from the home directory
        # so it requires an absolute path
        minpower_args["directory"] = os.path.abspath(minpower_args["directory"])

        if (
            "scenarios_directory" in minpower_args
            and minpower_args["scenarios_directory"]
        ):
            minpower_args["scenarios_directory"] = os.path.abspath(
                minpower_args["scenarios_directory"]
            )

    if args.verbose:
        pprint(minpower_args)

    scheduler_call = []
    minpower_call = ["minpower"]
    stdout = sys.stdout
    stderr = subprocess.STDOUT

    def arg2str(k, v):
        if k == "directory":
            s = v
        elif v == True:
            s = "--{k}".format(k=k)
        else:
            s = "--{k}={v}".format(k=k, v=v)
        return s

    # make a big chain of args
    minpower_call.extend(
        sorted(
            [
                arg2str(k, v)
                for k, v in list(minpower_args.items())
                if (k == "directory")
                or (
                    (k in default_minpower_config) and (v != default_minpower_config[k])
                )
            ]
        )
    )

    mode = args.scheduler_mode
    if mode == "pass":
        # just let all the commands pass through
        pass
    elif mode == "nohup":
        scheduler_call = ["nohup"]
        if args.dry_run:
            print(
                (
                    "would have opened files to write"
                    + "\n\tstdout: {p}.out\n\tstderr: {p}.err".format(p=os.getpid())
                )
            )
        else:
            stdout = open("{}.out".format(os.getpid()), "w")
            stderr = open("{}.err".format(os.getpid()), "w")
    elif mode == "qsub":
        # see https://sig.washington.edu/itsigs/Hyak_Job_Scheduler
        # default walltime limit is only one hour - must set this
        scheduler_call = [
            "qsub",
            "-l nodes=1:ppn=12,feature=12core,mem={m}gb,walltime={h}:00:00".format(
                m=args.memory, h=args.hours_limit
            ),
        ]

        if args.email:
            scheduler_call.extend(
                ["-m ae", "-M {e}".format(e=args.email)]  # mail on completion/failure
            )

        # need to write a script to disk to call with qsub
        if minpower_args["standalone_restart"]:
            # write to the same script, but comment out the original call
            script_name = "./{}.sh".format(minpower_args["pid"])
            if args.dry_run:
                print(
                    (
                        "would have commented out the original call in {}".format(
                            script_name
                        )
                    )
                )
            else:
                with open(script_name, "r") as f:
                    old_script = "\n".join(
                        ["# original call"]
                        + ["#" + ln for ln in f.readlines()]
                        + ["", ""]
                    )
                with open(script_name, "w") as f:
                    f.write(old_script)
            script_mode = "a"
        else:
            script_name = "./{}.sh".format(os.getpid())
            script_mode = "w+"
        if args.dry_run:
            print(
                (
                    "would have written script {f}: \n{c}".format(
                        f=script_name, c=" ".join(minpower_call)
                    )
                )
            )
        else:
            with open(script_name, script_mode) as f:
                f.write(" ".join(minpower_call))
        minpower_call = [script_name]

    # actually make the call
    if args.dry_run:
        print(
            (
                "would have executed as a {p}:\n{c}".format(
                    p="child process" if mode == "pass" else "subprocess",
                    c=" ".join((scheduler_call + minpower_call)),
                )
            )
        )
    else:
        if mode == "pass":
            subprocess.call(scheduler_call + minpower_call)
            pid = None
        else:

            pid = subprocess.Popen(
                scheduler_call + minpower_call,
                stdout=stdout,
                stderr=stderr,
            ).pid

    if args.verbose:
        print(("parent process {}".format(os.getpid())))
        print(("starting run {}".format(pid)))


if __name__ == "__main__":
    main()



"""
Power systems optimization problem solver
for ED, OPF, UC, and SCUC problems. The :mod:`solve`
module contains the top-level commands for creating
problems and solving them.
"""

import traceback
import sys
import os
import logging
import subprocess
import time as timer
import argparse
import pdb

from .config import user_config, parse_command_line_config
from .commonscripts import joindir, StreamToLogger
from . import powersystems
from . import get_data
from . import stochastic
from . import results
from .standalone import store_times, init_store, get_storage, repack_storage


def _set_store_filename(pid=None):
    fnm = "stage-store.hd5"
    if user_config.output_prefix or user_config.pid:
        fnm = "{}-{}".format(pid, fnm)

    user_config.store_filename = joindir(user_config.directory, fnm)


def solve_multistage_standalone(power_system, times, scenario_tree, data):

    stage_times = times.subdivide(
        user_config.hours_commitment, user_config.hours_overlap
    )

    pid = user_config.pid if user_config.pid else os.getpid()
    has_pid = user_config.output_prefix or user_config.pid

    if user_config.standalone_restart:
        # get the last stage in storage
        storage = get_storage()
        stg_start = len(storage["solve_time"].dropna())
        logging.info("Restarting on stage {}".format(stg_start))
        stage_times_remaining = stage_times[stg_start:]
    else:
        storage = init_store(power_system, stage_times, data)
        stage_times_remaining = stage_times
        stg_start = 0

    for stg, t_stage in enumerate(stage_times_remaining):
        logging.info("Stage starting at {}".format(t_stage.Start.date()))
        # store current stage times
        storage = store_times(t_stage, storage)
        storage.close()
        storage = None
        command = "standalone_minpower {dir} {stg} {pid} {db}".format(
            dir=user_config.directory,
            stg=stg + stg_start,
            pid="--pid {}".format(pid) if has_pid else "",
            db="--debugger" if user_config.debugger else "",
        )
        try:
            subprocess.check_call(command, shell=True, stdout=sys.stdout)
        except AttributeError:
            # HACK - avoid error when nose hijacks sys.stdout
            subprocess.check_call(command, shell=True)

    repack_storage()
    storage = get_storage()
    return storage, stage_times


def standaloneUC():
    """the hook for the ``standalone_minpower`` script"""
    from .standalone import store_state, load_state

    parser = argparse.ArgumentParser()
    parser.add_argument("directory", type=str, help="the problem direcory")
    parser.add_argument("stg", type=int, help="the stage number")
    parser.add_argument("--pid", default="", help="the process id of the parent")
    parser.add_argument(
        "--debugger", action="store_true", default=False, help="do some debugging"
    )

    args = parser.parse_args()
    stg = args.stg
    user_config.directory = args.directory
    user_config.pid = args.pid

    try:
        _set_store_filename(args.pid)
        # load stage data
        power_system, times, scenario_tree = load_state()

        # override the stored config with the current command line config
        user_config.directory = args.directory
        user_config.debugger = args.debugger

        _setup_logging(args.pid)

        sln = create_solve_problem(power_system, times, scenario_tree, stage_number=stg)

        store = store_state(power_system, times, sln)
        store.close()
    except:
        if user_config.debugger or args.debugger:
            __, __, tb = sys.exc_info()
            traceback.print_exc()
            pdb.post_mortem(tb)
        else:
            raise

    return


def solve_problem(
    datadir=".",
    shell=True,
    problemfile=False,
    csv=True,
):
    """
    Solve a optimization problem specified by spreadsheets in a directory.
    Read in data, create and solve the problem, and return solution.
    The problem type is determined by the data.
    All options are set within `user_config`.
    """
    user_config.directory = datadir

    pid = user_config.pid if user_config.pid else os.getpid()

    _set_store_filename(pid)
    _setup_logging(pid)

    if user_config.standalone_restart:
        store = get_storage()
        debugger = user_config.debugger  # preserve debugger state
        user_config.update(store["configuration"])
        user_config.debugger = debugger
        user_config.standalone_restart = True  # preserve restart state
        store.close()

    logging.debug(dict(user_config))

    start_time = timer.time()
    logging.debug("Minpower reading {}".format(datadir))
    generators, loads, lines, times, scenario_tree, data = get_data.parsedir()
    logging.debug("data read")
    power_system = powersystems.PowerSystem(generators, loads, lines)

    logging.debug("power system initialized")
    if times.spanhrs <= user_config.hours_commitment + user_config.hours_overlap:
        solution = create_solve_problem(power_system, times, scenario_tree)
    else:  # split into multiple stages and solve
        if user_config.standalone:
            stage_solutions, stage_times = solve_multistage_standalone(
                power_system, times, scenario_tree, data
            )
        else:
            stage_solutions, stage_times = solve_multistage(
                power_system, times, scenario_tree, data
            )
        solution = results.make_multistage_solution(
            power_system, stage_times, stage_solutions
        )

    if shell:
        if user_config.output_prefix or user_config.pid:
            stdout = sys.stdout
            sys.stdout = StreamToLogger()
            solution.show()
            sys.stdout = stdout
        solution.show()
    if csv and not user_config.standalone:
        solution.saveCSV()
    if user_config.visualization:
        solution.visualization()
    logging.info("total time: {}s".format(timer.time() - start_time))

    if user_config.on_complete_script:
        os.system(user_config.on_complete_script)

    return solution


def solve_multistage(power_system, times, scenario_tree=None, data=None):
    stage_times = times.subdivide(
        user_config.hours_commitment, user_config.hours_overlap
    )

    stage_solutions = []

    for stg, t_stage in enumerate(stage_times):
        logging.info("Stage starting at {}".format(t_stage.Start.date()))
        # solve
        solution = create_solve_problem(power_system, t_stage, scenario_tree, stg)
        # add to stage solutions
        stage_solutions.append(solution)
        # reset model
        power_system.reset_model()
        # set inital state for next stage
        if stg < len(stage_times) - 1:
            power_system.set_initialconditions(stage_times[stg + 1].initialTime)

    return stage_solutions, stage_times


def create_solve_problem(
    power_system, times, scenario_tree=None, stage_number=None, rerun=False
):
    """create and solve an optimization problem."""

    create_problem(power_system, times, scenario_tree, stage_number, rerun)

    instance = power_system.solve_problem(times)

    logging.debug("solved... get results")

    sln = results.make_solution(power_system, times)

    power_system.disallow_shedding()

    # resolve with observed power and fixed status
    if sln.is_stochastic:
        power_system.resolve_stochastic_with_observed(instance, sln)
    elif user_config.deterministic_solve or user_config.perfect_solve:
        power_system.resolve_determinisitc_with_observed(sln)

    if len(times) > 1:
        power_system.get_finalconditions(sln)

        power_system.disallow_shedding()
        sln.stage_number = stage_number

    return sln


def create_problem(
    power_system, times, scenario_tree=None, stage_number=None, rerun=False
):
    """Create an optimization problem."""

    logging.debug("initialized problem")
    power_system.create_variables(times)
    logging.debug("created variables")
    power_system.create_objective(times)
    logging.debug("created objective")
    power_system.create_constraints(times)
    logging.debug("created constraints")

    if scenario_tree is not None and sum(scenario_tree.shape) > 0 and not rerun:
        stochastic.construct_simple_scenario_tree(
            power_system, times, time_stage=stage_number
        )
        stochastic.define_stage_variables(power_system, times)
        stochastic.create_problem_with_scenarios(power_system, times)
    return


def _setup_logging(pid=None):
    """set up the logging to report on the status"""
    kwds = dict(
        level=int(user_config.logging_level)
        if not user_config.debugger
        else logging.DEBUG,
        datefmt="%Y-%m-%d %H:%M:%S",
        format="%(asctime)s %(levelname)s: %(message)s",
    )
    # log to file if pid is set, unless in debugging mode
    if (user_config.output_prefix or user_config.pid) and not user_config.debugger:
        kwds["filename"] = joindir(user_config.directory, "{}.log".format(pid))
    if (user_config.logging_level > 10) and ("filename" not in kwds):
        # don't log the time if debugging isn't turned on
        kwds["format"] = "%(levelname)s: %(message)s"
    logging.basicConfig(**kwds)


def main():
    """
    The command line interface for minpower. For more info use:
    ``minpower --help``
    """

    args = parse_command_line_config(
        argparse.ArgumentParser(description="Minpower command line interface")
    )

    directory = user_config.directory

    if not os.path.isdir(directory):
        msg = 'There is no folder named "{}".'.format(directory)
        raise OSError(msg)

    if args["profile"]:
        print("run profile")
        import cProfile

        prof = cProfile.Profile()
        prof.runcall(solve_problem, directory)
        prof.dump_stats("minpower.profile")

    else:
        # solve the problem with those arguments
        try:
            solve_problem(directory)
        except:
            if args["debugger"]:
                __, __, tb = sys.exc_info()
                traceback.print_exc()
                pdb.post_mortem(tb)
            else:
                raise


# for use in dev
if __name__ == "__main__":
    main()



"""
Get data from spreadsheet files and parse it into
:class:`~Generator`, :class:`~powersystems.Load`,
:class:`~powersystems.Bus`, and :class:`~powersystems.Line` objects.
Also extract the time information and create all
    :class:`~schedule.Timelist` objects.
"""
import pandas as pd
import xarray as xr
from pandas import DataFrame, Timestamp, read_csv
from glob import glob
from collections import OrderedDict

from . import powersystems
from .schedule import just_one_time, get_schedule, TimeIndex, make_constant_schedule
from .commonscripts import joindir, drop_case_spaces, set_trace

from .powersystems import PowerSystem
from .generators import Generator, Generator_Stochastic, Generator_nonControllable
from .config import user_config

import os
import logging

fields = dict(
    Line=["name", "tobus", "frombus", "reactance", "pmax"],
    Generator=[
        "name",
        "kind",
        "bus",
        "pmin",
        "pmax",
        "power",  # for a non-controllable gen in an ED
        "rampratemin",
        "rampratemax",
        "minuptime",
        "mindowntime",
        "startupramplimit",
        "shutdownramplimit",
        # for a bid points defined gen, noloadcost replaces the constant
        # polynomial
        "noloadcost",
        "costcurveequation",
        "heatrateequation",
        "fuelcost",
        "startupcost",
        "shutdowncost",
        "faststart",
        "mustrun",
        "sheddingallowed",
    ],
    Load=["name", "bus", "power"],
)

# extra fields for generator scheduling and bid specification
# these fields require parsing of additional files
# and result in additional objects being added
# to the Generator after creation
gen_extra_fields = [
    "observedname",
    "forecastname",
    "scenariosfilename",
    "scenariosdirectory",
    "costcurvepointsfilename",
]

fields_initial = ["status", "power", "hoursinstatus"]


def nice_names(df):
    """drop the case and spaces from all column names"""
    return df.rename(columns=dict([(col, drop_case_spaces(col)) for col in df.columns]))


def parse_standalone(storage, times):
    """load problem info from a pandas.HDFStore"""

    # filter timeseries data to only this stage
    timeseries = storage["data_timeseries"].loc[times.strings.values]

    # add loads
    loads = build_class_list(
        storage["data_loads"], powersystems.Load, times, timeseries
    )
    # add generators
    generators = build_class_list(
        storage["data_generators"], Generator, times, timeseries
    )
    # add lines
    lines = build_class_list(
        storage["data_lines"], powersystems.Line, times, timeseries
    )

    power_system = PowerSystem(generators, loads, lines)

    gen = power_system.get_generator_with_scenarios()
    if gen:
        # TODO - maybe only extract current day's values here
        scenario_values = storage["data_scenario_values"]
        gen.scenario_values = scenario_values
    else:
        scenario_values = xr.DataArray()

    return power_system, times, scenario_values


def _load_raw_data():
    """import data from spreadsheets"""
    datadir = user_config.directory

    if not os.path.isdir(datadir):
        raise OSError('data directory "{d}" does not exist'.format(d=datadir))
    [file_gens, file_loads, file_lines, file_init] = [
        joindir(datadir, filename)
        for filename in (
            user_config.file_gens,
            user_config.file_loads,
            user_config.file_lines,
            user_config.file_init,
        )
    ]

    generators_data = nice_names(read_csv(file_gens))
    loads_data = nice_names(read_csv(file_loads))

    try:
        lines_data = nice_names(read_csv(file_lines))
    except Exception:
        lines_data = DataFrame()

    try:
        init_data = nice_names(read_csv(file_init))
    except Exception:
        init_data = DataFrame()

    return generators_data, loads_data, lines_data, init_data


def _parse_raw_data(generators_data, loads_data, lines_data, init_data):
    # create times
    timeseries, times, generators_data, loads_data = setup_times(
        generators_data, loads_data
    )

    # add loads
    loads = build_class_list(loads_data, powersystems.Load, times, timeseries)

    # data modifiers
    if user_config.pmin_multiplier != 1.0:
        generators_data["pmin"] *= user_config.pmin_multiplier
        invalid_pmin = generators_data.pmin > generators_data.pmax
        if invalid_pmin.any():
            logging.warning(
                "Pmin scaling resulted in Pmin>Pmax. "
                + "Setting Pmin=Pmax for these {} generators.".format(
                    invalid_pmin.sum()
                )
            )
            generators_data.loc[invalid_pmin, "pmin"] = generators_data.loc[
                invalid_pmin, "pmax"
            ]

    if user_config.ramp_limit_multiplier != 1.0:
        generators_data["rampratemin"] *= user_config.ramp_limit_multiplier
        generators_data["rampratemax"] *= user_config.ramp_limit_multiplier

    if user_config.ignore_minhours_constraints:
        generators_data["minuptime"] = 0
        generators_data["mindowntime"] = 0
    if user_config.ignore_ramping_constraints:
        generators_data["rampratemax"] = None
        generators_data["rampratemin"] = None
    if user_config.ignore_pmin_constraints:
        generators_data["pmin"] = 0

    # add generators
    generators = build_class_list(generators_data, Generator, times, timeseries)

    # add lines
    lines = build_class_list(lines_data, powersystems.Line)
    # add initial conditions
    setup_initialcond(init_data, generators, times)

    # get scenario values (if applicable)
    scenario_values = setup_scenarios(generators_data, generators, times)

    # also return the raw DataFrame objects
    data = dict(
        generators=generators_data,
        loads=loads_data,
        lines=lines_data,
        init=init_data,
        timeseries=timeseries,
        scenario_values=scenario_values,
    )

    return generators, loads, lines, times, scenario_values, data


def parsedir(**filename_kwargs):
    """
    Import data from spreadsheets and build lists of
    :mod:`powersystems` classes.
    """

    generators_data, loads_data, lines_data, init_data = _load_raw_data(
        **filename_kwargs
    )

    return _parse_raw_data(generators_data, loads_data, lines_data, init_data)


def setup_initialcond(data, generators, times):
    """
    Take a list of initial conditions parameters and
    add information to each :class:`~Generator` object.
    """
    if len(times) <= 1:
        return  # for UC,ED no need to set initial status

    if len(data) == 0:
        logging.warning(
            """No generation initial conditions file found.
            Setting to defaults."""
        )
        for gen in generators:
            gen.set_initial_condition()
        return

    # begin by setting initial condition for all generators to off
    for g in generators:
        g.set_initial_condition(status=False, power=0)

    names = [g.name for g in generators]

    if not "name" in data.columns:
        # assume they are in order
        data["name"] = names

    if "power" not in data.columns:
        raise KeyError('initial conditions file should contain "power".')

    # add initial conditions for generators
    # which are specified in the initial file
    for i, row in data.iterrows():
        g = names.index(row["name"])
        kwds = row[fields_initial].dropna().to_dict()
        generators[g].set_initial_condition(**kwds)
    return


def build_class_list(data, model, times=None, timeseries=None):
    """
    Create list of class instances from the row of a DataFrame.
    """
    datadir = user_config.directory
    is_generator = model == Generator

    all_models = []

    if "schedulename" in data.columns:
        data["schedule"] = None

    for i, row in data.iterrows():
        row_model = model
        row = row.dropna()

        power = row.get("power")
        schedulename = row.get("schedulename")

        if is_generator:
            # get all those extra things which need to be parsed
            observed_name = row.get("observedname")
            forecast_name = row.get("forecastname")

            scenariosfilename = row.get("scenariosfilename")
            scenariosdirectory = row.get("scenariosdirectory")

            if scenariosdirectory and user_config.scenarios_directory:
                # override the scenarios directory with the one \
                # specified in the commandline options
                scenariosdirectory = user_config.scenarios_directory
                data.loc[i, "scenariosdirectory"] = user_config.scenarios_directory

            bid_points_filename = row.get("costcurvepointsfilename")

        if is_generator:
            if (
                schedulename
                or power
                or (forecast_name and user_config.deterministic_solve)
                or (observed_name and user_config.perfect_solve)
            ):
                row_model = Generator_nonControllable
            elif scenariosdirectory or scenariosfilename:
                row_model = Generator_Stochastic

        # warn about fields not in model
        valid_fields = pd.Index(fields[model.__name__] + ["schedulename"])
        if is_generator:
            valid_fields = valid_fields.union(pd.Index(gen_extra_fields))
        invalid_fields = row.index.difference(valid_fields)
        if len(invalid_fields) > 0:
            raise ValueError(
                "invalid fields in model:: {}".format(invalid_fields.tolist())
            )
            # logging.warning

        kwds = row[row.index.isin(fields[model.__name__])].to_dict()

        # add in any schedules
        if schedulename:
            kwds["schedule"] = timeseries[schedulename]
        elif pd.notnull(power):
            # a constant power schedule
            kwds["schedule"] = make_constant_schedule(times, power)
            kwds.pop("power")

        if is_generator:
            if observed_name:
                kwds["observed_values"] = timeseries[observed_name]

            if user_config.perfect_solve and observed_name:
                # for a perfect information solve forecast = observed
                kwds["schedule"] = kwds["observed_values"] = timeseries[observed_name]
            elif forecast_name:
                kwds["schedule"] = timeseries[forecast_name]

            if scenariosdirectory:
                try:
                    kwds["observed_values"] = timeseries[observed_name]
                except:
                    raise IOError(
                        """you must provide an
                        observed filename for a rolling stochastic UC"""
                    )

            # add a custom bid points file with {power, cost} columns
            if bid_points_filename:
                kwds["bid_points"] = read_bid_points(
                    joindir(datadir, bid_points_filename)
                )
                kwds["costcurveequation"] = None

        try:
            obj = row_model(index=i, **kwds)
        except TypeError:
            print(("{} model got unexpected parameter".format(model)))
            raise

        all_models.append(obj)

    return all_models


def read_bid_points(filename):
    bid_points = read_csv(filename)
    # return a dataframe of bidpoints
    return bid_points[["power", "cost"]].astype(float)


def setup_times(generators_data, loads_data):
    """
    Create a :class:`~schedule.TimeIndex` object
    from the schedule files.

    Also create a unified DataFrame of all the schedules, `timeseries`.

    If there are no schedule files (as in ED,OPF),
    create an index with just a single time.
    """
    fcol = "schedulefilename"
    ncol = "schedulename"

    loads_data[ncol] = None
    generators_data[ncol] = None

    if fcol not in loads_data.columns:
        loads_data[fcol] = None
    if fcol not in generators_data.columns:
        generators_data[fcol] = None

    datadir = user_config.directory

    timeseries = {}

    def filter_notnull(df, col):
        return df[df[col].notnull()]

    for i, load in filter_notnull(loads_data, fcol).iterrows():
        name = "d{}".format(i)
        loads_data.loc[i, ncol] = name
        timeseries[name] = (
            get_schedule(joindir(datadir, load[fcol])) * user_config.load_multiplier
            + user_config.load_adder
        )

    for i, gen in filter_notnull(generators_data, fcol).iterrows():
        name = "g{}".format(i)
        generators_data.loc[i, ncol] = name
        timeseries[name] = get_schedule(joindir(datadir, gen[fcol]))

    # handle observed and forecast power
    fobscol = "observedfilename"
    obscol = "observedname"
    ffcstcol = "forecastfilename"
    fcstcol = "forecastname"

    obs_name = None
    if fobscol in generators_data:
        generators_data[obscol] = None
        for i, gen in filter_notnull(generators_data, fobscol).iterrows():
            obs_name = "g{}_observations".format(i)
            generators_data.loc[i, obscol] = obs_name
            timeseries[obs_name] = get_schedule(joindir(datadir, gen[fobscol]))
            if user_config.wind_multiplier != 1.0:
                timeseries[obs_name] *= user_config.wind_multiplier

        generators_data = generators_data.drop(fobscol, axis=1)

    fcst_name = None
    if ffcstcol in generators_data:
        generators_data[fcstcol] = None
        for i, gen in filter_notnull(generators_data, ffcstcol).iterrows():
            fcst_name = "g{}_forecast".format(i)
            generators_data.loc[i, fcstcol] = fcst_name
            timeseries[fcst_name] = (
                get_schedule(joindir(datadir, gen[ffcstcol]))
                * user_config.wind_multiplier
                + user_config.wind_forecast_adder
            )

            if user_config.wind_error_multiplier != 1.0:
                logging.debug("scaling wind forecast error")
                obs_name = "g{}_observations".format(i)
                error = timeseries[fcst_name] - timeseries[obs_name]
                timeseries[fcst_name] = (
                    timeseries[obs_name] + error * user_config.wind_error_multiplier
                )

            if (timeseries[fcst_name] < 0).any():
                print((timeseries[fcst_name].describe()))
                logging.warning("Wind forecast must always be at least zero.")
                timeseries[fcst_name][timeseries[fcst_name] < 0] = 0

        generators_data = generators_data.drop(ffcstcol, axis=1)

    generators_data = generators_data.drop(fcol, axis=1)
    loads_data = loads_data.drop(fcol, axis=1)

    if len(timeseries) == 0:
        # this is a ED or OPF problem - only one time
        return DataFrame(), just_one_time(), generators_data, loads_data

    timeseries = DataFrame(timeseries)
    times = TimeIndex(timeseries.index)
    timeseries.index = times.strings.values

    if user_config.wind_capacity_factor != 0:
        if len(filter_notnull(generators_data, obscol)) != 1:
            raise NotImplementedError(
                "wind capacity factor only works with one wind generator"
            )

        all_loads = timeseries[
            [col for col in timeseries.columns if col.startswith("d")]
        ]

        capf_current = timeseries[obs_name].sum() / all_loads.sum(axis=1).sum()

        wind_mult = user_config.wind_capacity_factor / capf_current
        user_config.wind_multiplier = wind_mult

        logging.info(
            "scaling wind from a c.f. of {} to a c.f. of {}".format(
                capf_current, user_config.wind_capacity_factor
            )
        )
        timeseries[obs_name] *= wind_mult
        if fcst_name:
            timeseries[fcst_name] *= wind_mult

    return timeseries, times, generators_data, loads_data


def _parse_scenario_day(filename):
    logging.debug("reading scenarios from %s", filename)
    data = read_csv(filename, parse_dates=True, index_col=0)

    # select subset of scenarios
    Nscenarios = user_config.scenarios
    if Nscenarios:
        data = data[data.index < Nscenarios]
        data["probability"] = data["probability"] / sum(data["probability"])

    # return the data with probability column first
    return data[data.columns.drop("probability").insert(0, "probability")]


def setup_scenarios(gen_data, generators, times):

    col = "scenariosdirectory"
    scenario_values = xr.DataArray()
    if (
        user_config.deterministic_solve
        or user_config.perfect_solve
        or (not col in gen_data.columns)
    ):
        # a deterministic problem
        return scenario_values

    gen_params = gen_data[gen_data[col].notnull()]
    if len(gen_params) > 1:
        raise NotImplementedError("more than one generator with scenarios.")

    gen = generators[gen_params.index[0]]
    gen.has_scenarios = True

    # directory of scenario values where each file is one day
    scenarios_directory = gen_params["scenariosdirectory"].values[0]

    searchstr = "*.csv"

    filenames = sorted(
        glob(joindir(user_config.directory, joindir(scenarios_directory, searchstr)))
    )
    if not filenames:
        raise IOError('no scenario files in "{}"'.format(scenarios_directory))

    alldata = OrderedDict()
    for i, f in enumerate(filenames):
        data = _parse_scenario_day(f)
        # label scenarios for the day with the date
        date = Timestamp(data.columns.drop("probability")[0]).date()
        alldata[date] = data

    # TODO - assumes one hour intervals!!
    hrs = user_config.hours_commitment + user_config.hours_overlap

    # make scenarios into a xarray with axes: day, scenario, {prob, [hours]}
    # FIXME: make this work
    scenario_values = xr.DataArray(
        items=list(alldata.keys()),
        major_axis=list(range(max([len(dat) for dat in list(alldata.values())]))),
        minor_axis=["probability"] + list(range(hrs)),
    )

    for day, scenarios in list(alldata.items()):
        if "probability" == scenarios.columns[-1]:
            # reoder so that probability is the first column
            scenarios = scenarios[scenarios.columns[:-1].insert(0, "probability")]
        # rename the times into just hour offsets
        scenarios = scenarios.rename(
            columns=dict(
                list(
                    zip(
                        scenarios.columns,
                        ["probability"] + list(range(len(scenarios.columns) - 1)),
                    )
                )
            )
        )

        # and take the number of hours needed
        scenarios = scenarios[scenarios.columns[: 1 + hrs]]

        scenario_values[day] = scenarios

    if user_config.wind_multiplier != 1.0:
        # FIXME: make this work
        scenario_values *= user_config.wind_multiplier
        svt = scenario_values.transpose(2, 1, 0)
        svt["probability"] *= 1 / user_config.wind_multiplier
        scenario_values = svt.transpose(2, 1, 0)

    gen.scenario_values = scenario_values
    # defer scenario tree construction until actual time stage starts
    return scenario_values


def _has_valid_attr(obj, name):
    return getattr(obj, name, None) is not None



from pandas import Series, DataFrame
from minpower.config import user_config
from minpower import powersystems, schedule, solve
from minpower.generators import Generator
from minpower.optimization import value, OptimizationError

import nose
from nose.tools import istest, with_setup, raises, set_trace
from nose.tools import eq_ as assert_equal


import logging

logging.basicConfig(level=logging.CRITICAL, format="%(levelname)s: %(message)s")


default_config = user_config.copy()


def get_duals():
    user_config.duals = True


def reset_config():
    # note - can't set a new user_config, because it would be a local variable
    #   instead, we update it with the defaults
    #   this only works if we aren't adding anything extra to the user_config
    user_config.update(default_config)
    try:
        assert user_config == default_config
    except:
        print(
            (
                DataFrame(
                    dict(config=Series(user_config), default=Series(default_config))
                )
            )
        )
        raise


singletime = schedule.just_one_time()


gen_costs = dict(cheap=10, mid=20, expensive=30)


def assertAlmostEqual(x, y):
    assert round(x - y, 5) == 0


# def make_single_bus(generators,loads):
# singlebus=powersystems.Bus()
# singlebus.generators=generators
# singlebus.loads=loads
# return [singlebus]


def make_cheap_gen(**kwargs):
    return Generator(
        name="cheap gen", costcurveequation="{}P".format(gen_costs["cheap"]), **kwargs
    )


def make_mid_gen(**kwargs):
    return Generator(
        name="middle-range gen",
        costcurveequation="{}P".format(gen_costs["mid"]),
        **kwargs
    )


def make_expensive_gen(**kwargs):
    if "costcurveequation" not in kwargs:
        kwargs["costcurveequation"] = "{}P".format(gen_costs["expensive"])
    return Generator(name="expensive gen", **kwargs)


def make_loads_times(Pd=200, Pdt=None, **kwargs):
    if Pdt is None:
        times = singletime
        sched = Series(Pd, index=times)
        loads = [powersystems.Load(schedule=sched, **kwargs)]
    else:
        times = schedule.make_times_basic(N=len(Pdt))
        sched = Series(Pdt, index=times)
        loads = [powersystems.Load(schedule=sched, **kwargs)]

    return dict(loads=loads, times=times)


def solve_problem(
    generators, loads=None, times=None, gen_init=None, lines=None, do_reset_config=True
):

    if lines is None:
        lines = []

    if len(times) > 0:
        for g, gen in enumerate(generators):
            gen.index = g
            if gen_init is None:
                gen.set_initial_condition()
            else:
                gen.set_initial_condition(**gen_init[g])

    power_system = powersystems.PowerSystem(generators, loads, lines)
    solve.create_problem(power_system, times)

    try:
        instance = power_system.solve()
    except OptimizationError:
        # re-do stage, with load shedding allowed
        logging.critical("stage infeasible, re-running with load shedding.")
        power_system.allow_shedding(times)
        try:
            instance = power_system.solve()
        except OptimizationError:
            scheduled, committed = power_system.debug_infeasibe(times)
            power_system.write_model("infeasible.lp")
            raise OptimizationError("failed to solve, even with load shedding.")

    if do_reset_config:
        reset_config()

    if power_system.solved:
        power_system.update_variables()
    else:
        power_system.write_model("problem.lp")
        raise OptimizationError("infeasible problem, wrote to problem.lp")
    return power_system, times



"""
Do an many economic dispatches over a range of load values to get an
'aggregate' cost curve for a system. 
"""


from minpower.config import user_config
from minpower.get_data import parsedir
from minpower.commonscripts import joindir
from minpower.tests.test_utils import make_loads_times, solve_problem
from pandas import DataFrame, Series, Index
import numpy as np
import matplotlib.pyplot as plt
import argparse


def main(args):
    generators, loads, _, times, _, data = parsedir()
    generators = [gen for gen in generators if gen.is_controllable]

    gen_data = data["generators"]
    if args["min"] == 0:
        args["min"] = 1.1 * gen_data.pmin.sum()

    if args["max"] == 0:
        args["max"] = 0.99 * gen_data.pmax.sum()

    load_values = np.arange(args["min"], args["max"], args["interval"])
    results = DataFrame(
        columns=["prices", "committed", "last_committed"], index=load_values
    )

    committed_gen_names = Index([])

    for load_val in load_values:
        print(load_val)
        loads_times = make_loads_times(Pd=load_val)
        power_system, times = solve_problem(
            generators, do_reset_config=False, **loads_times
        )
        t = times[0]
        results.loc[load_val, "prices"] = power_system.buses[0].price(t)
        statuses = Series(
            dict([(gen.name, gen.status(t).value) for gen in power_system.generators()])
        )

        results.loc[load_val, "committed"] = statuses.sum()
        results.loc[load_val, "last_committed"] = statuses[
            statuses == 1
        ].index.difference(committed_gen_names)
        committed_gen_names = statuses[statuses == 1].index

    if (load_values[-1] == 0.99 * gen_data.pmax.sum()) and (
        statuses.sum() != len(generators)
    ):
        print("warning: uncommitted generation:")
        print((gen_data.set_index("name").loc[statuses[statuses == 0].index]))

    results.to_csv(joindir(user_config.directory, "ed_sweep.csv"))

    if args["hide_units_committed"]:
        ax = results.prices.plot(drawstyle="steps")
    else:
        ax = results[["prices", "committed"]].plot(
            drawstyle="steps", secondary_y=["committed"]
        )
        ax.right_ax.set_ylabel("Units committed")

    ax.set_xlabel("System Load (MW)")
    ax.set_ylabel("Estimated System Price ($/MWh)")

    plt.savefig(joindir(user_config.directory, "ed_sweep.png"))


def get_args():
    parser = argparse.ArgumentParser(
        description="""
        Do an many economic dispatches over a range of load values to get an
        'aggregate' cost curve for a system. """
    )
    parser.add_argument(
        "directory", type=str, help="the direcory of the problem you want to solve"
    )
    parser.add_argument("--min", type=int, default=0, help="the load value to start at")
    parser.add_argument("--max", type=int, default=0, help="the load value to end at")
    parser.add_argument(
        "--interval",
        type=float,
        default=100,
        help="the interval to increment the load by",
    )
    parser.add_argument(
        "--hide_units_committed", "-c", action="store_true", default=False
    )

    args = parser.parse_args()

    user_config.directory = args.directory
    user_config.duals = True
    user_config.perfect_solve = True  # hack - ignore scenarios
    user_config.dispatch_decommit_allowed = True

    return vars(args)


if __name__ == "__main__":
    args = get_args()
    main(args)



"""
A module to workaround a memory leak in coopr by saving the results
of each day of a rolling unit commitment to disk (in HDF format)
and reloading them to run the next day as a memory independent subprocess.
"""

import os
import logging
import sys
import pandas as pd
from pandas import Series, DataFrame
from .commonscripts import (
    gen_time_dataframe,
    correct_status,
    debug_frame_unequal,
    set_trace,
)
from .config import user_config

from .schedule import TimeIndex
from .get_data import parse_standalone
import pkg_resources


def wipe_storage():
    try:
        os.remove(user_config.store_filename)
        logging.debug("wipe storage")
    except OSError:
        pass


def get_storage():
    return pd.HDFStore(user_config.store_filename)


def store_times(tstage, storage=None):
    if storage is None:
        storage = get_storage()
    storage["times"] = tstage.strings
    return storage


def init_store(power_system, times, data):
    """create the store before the first stage"""
    wipe_storage()
    storage = get_storage()
    generators = power_system.generators()

    # store the problem info read from the spreadsheets
    for key, df in list(data.items()):
        if key != "scenario_values":
            for k, v in (df.dtypes == object).iterkv():
                if v:
                    df[k] = df[k].fillna("")

        key = "data_" + key
        storage[key] = df

    stages = len(times)  # stage_times
    est_Nt = len(data["timeseries"])
    t = [times[0].initialTime]

    # store first stage initial state
    storage.append(
        "power",
        gen_time_dataframe(
            generators, t, values=[[gen.initial_power for gen in generators]]
        ),
        expectedrows=est_Nt,
    )

    storage.append(
        "status",
        gen_time_dataframe(
            generators, t, values=[[gen.initial_status for gen in generators]]
        ),
        expectedrows=est_Nt,
    )

    storage["hrsinstatus"] = gen_time_dataframe(
        generators, t, values=[[gen.initial_status_hours for gen in generators]]
    )

    # setup empty containers for variables
    storage["load_shed"] = Series()
    storage["gen_shed"] = Series()
    storage["expected_cost"] = DataFrame()
    storage["observed_cost"] = DataFrame()
    storage["expected_status"] = DataFrame()
    storage["expected_power"] = DataFrame()
    storage["expected_fuelcost"] = DataFrame()
    storage["observed_fuelcost"] = DataFrame()

    # setup one-per-stage results
    storage["solve_time"] = Series(index=list(range(stages)))
    storage["mipgap"] = Series(index=list(range(stages)))

    # store configuration
    storage["configuration"] = Series(user_config)

    storage["version"] = Series(
        {
            "minpower": pkg_resources.get_distribution("minpower").version,
            # try storing version number of the current working directory
            "problem": _get_problem_version(),
        }
    )

    return storage


def _get_problem_version():
    version = ""
    possible_init = os.path.join(os.path.abspath(user_config.directory), "__init__.py")
    if os.path.exists(possible_init):
        vstr = open(possible_init, "r").read()
        version = vstr.split('"')[1].rstrip('"')
    return version


def store_state(power_system, times, sln=None):
    storage = get_storage()
    generators = power_system.generators()

    stg = sln.stage_number
    table_append(storage, "power", sln.generators_power)
    table_append(storage, "status", sln.generators_status)
    table_append(storage, "load_shed", sln.load_shed_timeseries)
    table_append(storage, "gen_shed", sln.gen_shed_timeseries)

    tEnd = times.last_non_overlap()
    storage["hrsinstatus"] = gen_time_dataframe(
        generators,
        [tEnd],
        values=[[gen.finalstatus["hoursinstatus"] for gen in generators]],
    )

    _add_tbl_val(storage, "solve_time", stg, sln.solve_time)
    _add_tbl_val(storage, "mipgap", stg, sln.mipgap)

    if sln._resolved:
        table_append(storage, "observed_cost", sln.observed_totalcost)
        table_append(storage, "observed_fuelcost", sln.observed_fuelcost)
        table_append(storage, "expected_cost", sln.expected_totalcost)
        table_append(storage, "expected_fuelcost", sln.expected_fuelcost)
        table_append(storage, "expected_power", sln.expected_power)
        table_append(storage, "expected_status", sln.expected_status)
    else:
        table_append(storage, "expected_cost", sln.totalcost_generation)
        table_append(storage, "expected_fuelcost", sln.fuelcost)

    #    # DEBUGGING
    #    from pandas.util.testing import assert_frame_equal
    #    try:
    #        gens = storage['data_generators'].copy()
    #        gens.index = ['g{}'.format(g) for g in gens.index]
    #        status = storage['status'].loc[1:]
    #        power = storage['power'][status.columns].loc[status.index]
    #        cost = storage['observed_cost'][status.columns].loc[status.index]
    #        fuelcost = storage['observed_fuelcost'][status.columns].loc[status.index]
    #        stcost = cost - fuelcost

    #
    #        pcheck = (power * status)
    #        ccheck = (cost * status)
    #        stcheck = (stcost * status)
    #        assert_frame_equal(pcheck, power)
    #        assert_frame_equal(ccheck, cost)
    #        assert_frame_equal(stcheck, stcost)
    #        assert(((stcost < gens.startupcost - 1e-5) & (stcost > 1e-5) ).sum().sum() == 0)
    #
    #    except AssertionError:
    #        raise

    return storage


def load_state():
    storage = get_storage()
    user_config.update(storage["configuration"].to_dict())

    # set up the times
    startidx = int(storage["times"][0].strip("t"))
    times = TimeIndex(storage["times"].index, startidx)
    intervalhrs = (
        times.strings.index[1] - times.strings.index[0]
    ).total_seconds() / 3600.0
    times._int_division = user_config.hours_commitment / intervalhrs
    times._int_overlap = user_config.hours_overlap / intervalhrs
    if len(times) <= times._int_division:
        # dont set overlap for last stage
        times._int_overlap = 0

    # create power_system
    power_system, times, scenario_tree = parse_standalone(storage, times)
    generators = power_system.generators()

    # set up initial state
    t = times.initialTime
    status = correct_status(storage["status"]).loc[t]
    for gen in generators:
        g = str(gen)
        gen.set_initial_condition(
            power=storage["power"][g][t],
            status=status[g],
            hoursinstatus=storage["hrsinstatus"][g][t],
        )

    return power_system, times, scenario_tree


def repack_storage():
    """compress that ballooning storage"""
    # http://stackoverflow.com/questions/13089359/mystery-when-storing-a
    # -dataframe-containing-strings-in-hdf-with-pandas
    os.system(
        "ptrepack {f} {f}.copy; mv {f}.copy {f};".format(f=user_config.store_filename)
    )


def _add_tbl_val(storage, tablename, index, value):
    tbl = storage[tablename]
    tbl[index] = value
    storage[tablename] = tbl


def table_append(store, name, newvals):
    store[name] = store[name].append(newvals)
    return



"""Test the all of the solver links"""

from minpower import optimization, config
from .test_utils import *
from pyomo import environ as pyomo


mem_tracking = False
if mem_tracking:
    from pympler.classtracker import ClassTracker
    from pympler.classtracker_stats import HtmlStats
    import objgraph

    tracker = ClassTracker()
    for cls in [
        pyomo.ConcreteModel,
        pyomo.Var,
        pyomo.base.var._VarElement,
        pyomo.Constraint,
        optimization.OptimizationProblem,
    ]:
        tracker.track_class(cls)


def simple_problem():
    prob = optimization.OptimizationProblem()
    if mem_tracking:
        tracker.create_snapshot("prob. init")
    prob.add_variable("x", low=0, high=3)
    prob.add_variable("y", low=0, high=1)
    x = prob.get_component("x")
    y = prob.get_component("y")
    prob.add_objective(y - 4 * x)
    prob.add_constraint("ineq", x + y <= 2)
    if mem_tracking:
        tracker.create_snapshot("prob. created")
    return prob


def run_one_solver(solver_name):
    prob = simple_problem()
    orig_solver = config.user_config.solver
    config.user_config.duals = False
    config.user_config.solver = solver_name

    prob.solve()
    if mem_tracking:
        tracker.create_snapshot("prob. solved")
    status = prob.solved
    del prob
    if mem_tracking:
        tracker.create_snapshot("prob. deleted")
    if solver_name == "glpk" and mem_tracking:
        tracker.stats.print_summary()
        # HtmlStats(tracker=tracker).create_html('profile-simple-problem.html')
        models_left = objgraph.by_type("ConcreteModel")
        print(models_left)
        objgraph.show_backrefs(models_left, filename="backrefs-simple-problem.png")

    config.user_config.duals = True
    config.user_config.solver = orig_solver
    return status


@istest
def cplex():
    """Test each available solver on a simple problem"""
    if "cplex" in config.available_solvers:
        assert run_one_solver("cplex")


@istest
def glpk():
    """Test the glpk solver on a simple problem"""
    if "glpk" in config.available_solvers:
        assert run_one_solver("glpk")


@istest
def cbc():
    """Test the cbc solver on a simple problem"""
    if "cbc" in config.available_solvers:
        assert run_one_solver("cbc")


@istest
def gurobi():
    """Test the gurobi solver on a simple problem"""
    if "gurobi" in config.available_solvers:
        assert run_one_solver("gurobi")



import pandas as pd
import logging
from .config import user_config
from .commonscripts import update_attributes, bool_to_int

from .optimization import value, OptimizationObject
from .schedule import is_init
from . import bidding


class Generator(OptimizationObject):

    """
    A generator model.

    :param pmin: minimum real power
    :param pmax: maximum real power
    :param minuptime: min. time after commitment in on status (hours)
    :param mindowntime: min. time after de-commitment in off status (hours)
    :param rampratemax: max. positive change in real power over 1hr (MW/hr)
    :param rampratemin: max. negative change in real power over 1hr (MW/hr)
    :param startupramplimit: max. positive change in real power over the
        first hour after startup (MW/hr)
    :param shutdownramplimit: max. negative change in real power over the
        last hour before shutdown (MW/hr)
    :param costcurveequation: text describing a polynomial cost curve ($/MWh)
      see :meth:`~bidding.parsePolynomial` for more.
    :param heatratestring: text describing a polynomial heat rate curve (MBTU/MW).
        converts to cost curve when multiplied by fuelcost.
    :param fuelcost: cost of fuel ($/MBTU)
    :param startupcost: cost to commit ($)
    :param shutdowncost: cost to de-commit ($)
    :param mustrun: flag that forces commimtent to be on

    :param name: name of the generator
    :param index: numbering of the generator
    :param bus: bus name that the generator is connected to
    """

    def __init__(
        self,
        kind="generic",
        pmin=0,
        pmax=500,
        minuptime=0,
        mindowntime=0,
        rampratemax=None,
        rampratemin=None,
        costcurveequation="20P",
        heatrateequation=None,
        fuelcost=1,
        bid_points=None,
        noloadcost=0,
        startupcost=0,
        shutdowncost=0,
        startupramplimit=None,
        shutdownramplimit=None,
        faststart=False,
        mustrun=False,
        name="",
        index=None,
        bus=None,
    ):

        update_attributes(self, locals())  # load in inputs

        # The formulation below requires that startup ramp limits are set.
        # The defaults are the normal ramp rate limits,
        # unless Pmin so high and Rmin/max small that startup is infeasible.
        # In that case the default is Pmin.

        if (self.startupramplimit is None) and (self.rampratemax is not None):
            if self.pmin > self.rampratemax:
                self.startupramplimit = self.pmin
            else:
                self.startupramplimit = self.rampratemax

        if (self.shutdownramplimit is None) and (self.rampratemin is not None):
            if -1 * self.pmin < self.rampratemin:
                self.shutdownramplimit = -1 * self.pmin
            else:
                self.shutdownramplimit = self.rampratemin

        self.fuelcost = float(fuelcost)

        self.is_controllable = True
        self.is_stochastic = False
        self.commitment_problem = True
        self.build_cost_model()
        self.init_optimization()

    def power(self, time=None, scenario=None):
        """real power output at time"""
        if time is not None and is_init(time):
            return self.initial_power
        else:
            return self.get_variable("power", time, scenario=scenario, indexed=True)

    def power_available(self, time=None, scenario=None):
        """power availble (constrained by pmax, ramprate, ...) at time"""
        if time is not None and is_init(time):
            return self.initial_power

        var_name = (
            "power_available"
            if self.commitment_problem and self.reserve_required
            else "power"
        )
        return self.get_variable(var_name, time, scenario=scenario, indexed=True)

    def status(self, time=None, scenario=None):
        """on/off status at time"""
        if self.commitment_problem or user_config.dispatch_decommit_allowed:
            if time is not None and is_init(time):
                return self.initial_status
            else:
                return self.get_variable(
                    "status", time, scenario=scenario, indexed=True
                )
        else:
            return 1

    def status_change(self, t, times):
        """is the unit changing status between t and t-1"""
        if t > 0:
            previous_status = self.status(times[t - 1])
        else:
            previous_status = self.initial_status
        return self.status(times[t]) - previous_status

    def cost(self, time, scenario=None, evaluate=False):
        """total cost at time (operating + startup + shutdown)"""
        return (
            self.operatingcost(time, scenario, evaluate)
            + self.cost_startup(time, scenario, evaluate)
            + self.cost_shutdown(time, scenario, evaluate)
        )

    def cost_startup(self, time, scenario=None, evaluate=False):
        if self.startupcost == 0 or not self.commitment_problem:
            return 0
        else:
            c = self.get_variable("startupcost", time, scenario=scenario, indexed=True)
            return c if not evaluate else value(c)

    def cost_shutdown(self, time, scenario=None, evaluate=False):
        if self.shutdowncost == 0 or not self.commitment_problem:
            return 0
        else:
            c = self.get_variable("shutdowncost", time, scenario=scenario, indexed=True)
            return c if not evaluate else value(c)

    def operatingcost(self, time=None, scenario=None, evaluate=False):
        """cost of real power production at time (based on bid model approximation)."""
        return self.bids.output(time, scenario=scenario, evaluate=evaluate)

    def truecost(self, time, scenario=None):
        """exact cost of real power production at time (based on exact bid polynomial)."""
        return value(self.status(time, scenario)) * self.bids.output_true(
            self.power(time, scenario)
        )

    def incrementalcost(self, time, scenario=None):
        """change in cost with change in power at time (based on exact bid polynomial)."""
        return (
            self.bids.output_incremental(self.power(time, scenario))
            if value(self.status(time, scenario))
            else None
        )

    def cost_first_stage(self, times):
        return sum(self.cost_startup(time) + self.cost_shutdown(time) for time in times)

    def cost_second_stage(self, times):
        return sum(self.operatingcost(time) for time in times)

    def getstatus(self, tend, times, status):
        return dict(
            status=value(self.status(tend)),
            power=value(self.power(tend)),
            hoursinstatus=self.gethrsinstatus(times, status),
        )

    def gethrsinstatus(self, times, stat):
        if not self.is_controllable:
            return 0

        end_status = stat.loc[stat.index[-1]]

        if (stat == end_status).all():
            intervals = len(stat)
            hrs = intervals * times.intervalhrs
            if self.initial_status == end_status:
                hrs += self.initial_status_hours
        else:
            noneq = stat[stat != end_status]
            if len(noneq) == 0:
                intervals = 0
            else:
                intervals = len(stat.loc[noneq.index[-1] :]) - 1

            hrs = intervals * times.intervalhrs

        return hrs

    def set_initial_condition(self, power=None, status=True, hoursinstatus=100):
        if power is None:
            # set default power as mean output
            power = (self.pmax - self.pmin) / 2
        if pd.isnull(power):
            raise ValueError("inital power cannot be null")
        self.initial_status = bool_to_int(status)
        self.initial_power = float(
            power * self.initial_status
        )  # note: this eliminates ambiguity of off status with power non-zero output
        self.initial_status_hours = hoursinstatus

    def build_cost_model(self):
        """
        parse the coefficients for the polynomial bid curve
        or custom bid points definition
        """
        bid_params = dict(
            owner=self,
            input_variable=self.power,
            min_input=self.pmin,
            max_input=self.pmax,
            status_variable=self.status,
        )

        if self.bid_points is None:
            # polynomial specification
            self.cost_breakpoints = user_config.breakpoints
            if getattr(self, "heatrateequation", None):
                self.cost_coeffs = [
                    self.fuelcost * coef
                    for coef in bidding.parse_polynomial(self.heatrateequation)
                ]
            else:
                self.cost_coeffs = bidding.parse_polynomial(self.costcurveequation)

            bid_params["polynomial"] = self.cost_coeffs
            bid_params["constant_term"] = self.cost_coeffs[0]
            bid_params["num_breakpoints"] = self.cost_breakpoints
            if self.noloadcost != 0:
                raise ValueError(
                    "no load cost should be defined as part of the polynomial."
                )

        else:
            # do some simple validation and delay construction to bidding
            # object
            min_power_bid = self.bid_points.power.min()
            max_power_bid = self.bid_points.power.max()

            if min_power_bid > self.pmin:  # pragma: no cover
                self.pmin = min_power_bid
                logging.warning(
                    "{g} should have a min. power bid ({mpb}) <= to its min. power limit ({mpl})".format(
                        g=str(self), mpb=min_power_bid, mpl=self.pmin
                    )
                )

            if max_power_bid < self.pmax:  # pragma: no cover
                self.pmax = max_power_bid
                logging.warning(
                    "{g} should have a max. power bid ({mpb}) >= to its max. power limit ({mpl})".format(
                        g=str(self), mpb=max_power_bid, mpl=self.pmax
                    )
                )

            bid_params["polynomial"] = None
            bid_params["bid_points"] = self.bid_points
            bid_params["constant_term"] = self.noloadcost

        self.bid_params = bid_params

    def create_variables(self, times):
        """
        Create the optimization variables for a generator over all times.
        Also create the :class:`bidding.Bid` objects and their variables.
        """
        self.commitment_problem = len(times) > 1
        self.add_variable("power", index=times.set, low=0, high=self.pmax)

        if self.commitment_problem or user_config.dispatch_decommit_allowed:
            self.add_variable(
                "status",
                index=times.set,
                kind="Binary",
                fixed_value=1 if self.mustrun else None,
            )

        if self.commitment_problem:
            # power_available exists for easier reserve requirement
            self.reserve_required = self._parent_problem().reserve_required
            if self.reserve_required:
                self.add_variable(
                    "power_available", index=times.set, low=0, high=self.pmax
                )
            if self.startupcost > 0:
                self.add_variable(
                    "startupcost", index=times.set, low=0, high=self.startupcost
                )
            if self.shutdowncost > 0:
                self.add_variable(
                    "shutdowncost", index=times.set, low=0, high=self.shutdowncost
                )

        self.bids = bidding.Bid(times=times, **self.bid_params)
        return

    def create_objective(self, times):
        return sum(self.cost(time) for time in times)

    def create_constraints(self, times):
        """create the optimization constraints for a generator over all times"""

        def roundoff(n):
            m = int(n)
            if n != m:  # pragma: no cover
                raise ValueError(
                    "min up/down times must be integer number of intervals, not {}".format(
                        n
                    )
                )
            return m

        if self.commitment_problem:
            # set initial and final time constraints
            tInitial = times.initialTimestr
            tEnd = len(times)
            if self.minuptime > 0:
                up_intervals_remaining = roundoff(
                    (self.minuptime - self.initial_status_hours) / times.intervalhrs
                )
                min_up_intervals_remaining_init = int(
                    min(tEnd, up_intervals_remaining * self.initial_status)
                )
            else:
                min_up_intervals_remaining_init = 0
            if self.mindowntime > 0:
                down_intervals_remaining = roundoff(
                    (self.mindowntime - self.initial_status_hours) / times.intervalhrs
                )
                min_down_intervals_remaining_init = int(
                    min(tEnd, down_intervals_remaining * (self.initial_status == 0))
                )
            else:
                min_down_intervals_remaining_init = 0
            # initial up down time
            if min_up_intervals_remaining_init > 0:
                self.add_constraint(
                    "minuptime",
                    tInitial,
                    0
                    >= sum(
                        [
                            (1 - self.status(times[t]))
                            for t in range(min_up_intervals_remaining_init)
                        ]
                    ),
                )
            if min_down_intervals_remaining_init > 0:
                self.add_constraint(
                    "mindowntime",
                    tInitial,
                    0
                    == sum(
                        [
                            self.status(times[t])
                            for t in range(min_down_intervals_remaining_init)
                        ]
                    ),
                )

            # initial ramp rate
            if self.rampratemax is not None:
                if self.initial_power + self.rampratemax < self.pmax:
                    E = self.power(times[0]) - self.initial_power <= self.rampratemax
                    self.add_constraint("ramp lim high", tInitial, E)

            if self.rampratemin is not None:
                if self.initial_power + self.rampratemin > self.pmin:
                    E = self.rampratemin <= self.power(times[0]) - self.initial_power
                    self.add_constraint("ramp lim low", tInitial, E)

            # calculate up down intervals
            min_up_intervals = roundoff(self.minuptime / times.intervalhrs)
            min_down_intervals = roundoff(self.mindowntime / times.intervalhrs)

            # reserve
            if self.reserve_required:

                def reserve_req(model, t):
                    return self.power(t) <= self.power_available(t)

                self.add_constraint_set("max gen power avail", times.set, reserve_req)

            # ramping power
            if self.rampratemax is not None:

                def ramp_max(model, t):
                    tPrev = get_tPrev(t, model, times)
                    ramp_limit = self.rampratemax * self.status(tPrev)
                    if self.startupramplimit is not None:
                        ramp_limit += self.startupramplimit * (
                            self.status(t) - self.status(tPrev)
                        )
                        # + self.pmax * (1 - self.status(times[t]))
                    return self.power_available(t) - self.power(tPrev) <= ramp_limit

                self.add_constraint_set("ramp limit high", times.set, ramp_max)

            #               # EQ19 from Carrion and Arroyo - has a conflicting
            #               # definition of shutdown power, available after shutdown hour?
            #                if self.shutdownramplimit is not None:
            #                    # if shutting down, available power must be reduced
            #                    self.add_constraint('ramp lim high shutdown', time,
            #                        self.power_available(times[t-1]) <= self.pmax * self.status(time) + \
            #                        self.shutdownramplimit * -1 * self.status_change(t, times)
            #                        )

            if self.rampratemin is not None:

                def ramp_min(model, t):
                    tPrev = get_tPrev(t, model, times)
                    ramp_limit = self.rampratemin * self.status(t)
                    # + self.pmax * (1 - self.status(times[t-1]))
                    if self.shutdownramplimit is not None:
                        ramp_limit += self.shutdownramplimit * (
                            -1 * (self.status(t) - self.status(tPrev))
                        )

                    return ramp_limit <= self.power_available(t) - self.power(tPrev)

                self.add_constraint_set("ramp limit low", times.set, ramp_min)

            # start up and shut down costs
            if self.startupcost > 0:

                def startupcostmin(model, t):
                    tPrev = get_tPrev(t, model, times)
                    return self.cost_startup(t) >= self.startupcost * (
                        self.status(t) - self.status(tPrev)
                    )

                self.add_constraint_set("startup cost min", times.set, startupcostmin)

                # these tightening constraints make stochastic problems take a very long time
            #                def startupcostmax(model, t):
            #                    return self.cost_startup(t) <= self.startupcost * self.status(t)
            #                self.add_constraint_set('startup cost max', times.set, startupcostmax)
            #
            #                def startupcostmax_prev(model, t):
            #                    tPrev = get_tPrev(t, model, times)
            #                    return self.cost_startup(t) <= self.startupcost * (1 - self.status(tPrev))
            #                self.add_constraint_set('startup cost max prev', times.set, startupcostmax_prev)

            if self.shutdowncost > 0:

                def shutdowncost(model, t):
                    tPrev = get_tPrev(t, model, times)
                    return self.cost_shutdown(t) >= self.shutdowncost * -1 * (
                        self.status(t) - self.status(tPrev)
                    )

                self.add_constraint_set("shutdown cost", times.set, shutdowncost)

            # note: costs must be >= constraints
            # for very large problems with >= constraints,
            # the solver may not always force the costs down to their minimum
            # however, the == formulation is incorrect and will not allow
            # unit shutdowns if the unit has a startup cost
            # solution is to use the min and max constraints together.

            # TODO: convert these to constraint list form
            for t, time in enumerate(times):

                # min up time
                if t >= min_up_intervals_remaining_init and self.minuptime > 0:
                    no_shut_down = list(range(t, min(tEnd, t + min_up_intervals)))
                    min_up_intervals_remaining = min(tEnd - t, min_up_intervals)
                    E = sum(
                        [self.status(times[s]) for s in no_shut_down]
                    ) >= min_up_intervals_remaining * self.status_change(t, times)
                    self.add_constraint("min up time", time, E)
                # min down time
                if t >= min_down_intervals_remaining_init and self.mindowntime > 0:
                    no_start_up = list(range(t, min(tEnd, t + min_down_intervals)))
                    min_down_intervals_remaining = min(tEnd - t, min_down_intervals)
                    E = sum(
                        [1 - self.status(times[s]) for s in no_start_up]
                    ) >= min_down_intervals_remaining * -1 * self.status_change(
                        t, times
                    )
                    self.add_constraint("min down time", time, E)

        # min/max power limits
        # these always apply (even if not a UC problem)
        if self.pmin > 0:

            def min_power(model, t):
                return self.power(t) >= self.status(t) * self.pmin

            self.add_constraint_set("min gen power", times.set, min_power)

        def max_power(model, t):
            return self.power_available(t) <= self.status(t) * self.pmax

        self.add_constraint_set("max gen power", times.set, max_power)

        return

    def __str__(self):
        return "g{ind}".format(ind=self.index)


def get_tPrev(t, model, times):
    return model.times.prev(t) if t != model.times.first() else times.initialTime


class Generator_nonControllable(Generator):

    """
    A generator with a fixed schedule.
    """

    def __init__(
        self,
        schedule=None,
        fuelcost=1,
        costcurveequation="0",
        bid_points=None,
        noloadcost=0,
        mustrun=False,
        faststart=False,
        sheddingallowed=False,
        pmin=0,
        pmax=None,
        name="",
        index=None,
        bus=None,
        kind="wind",
        observed_values=None,
        **kwargs
    ):

        update_attributes(self, locals())  # load in inputs
        self.is_controllable = False
        self.startupcost = 0
        self.shutdowncost = 0
        self.fuelcost = float(fuelcost)
        self.build_cost_model()
        self.init_optimization()
        self.is_stochastic = False
        self.shedding_mode = sheddingallowed and user_config.economic_wind_shed

    def power(self, time, scenario=None):
        if self.shedding_mode:
            power = self.get_variable(
                "power_used", time, scenario=scenario, indexed=True
            )
        else:
            power = self.power_available(time)
        return power

    def status(self, time=None, scenarios=None):
        return True

    def power_available(self, time=None, scenario=None):
        return self.get_parameter("power", time, indexed=True)

    def shed(self, time, scenario=None, evaluate=False):
        Pused = self.power(time, scenario=scenario)
        Pavail = self.power_available(time, scenario=scenario)
        if evaluate:
            Pused = value(Pused)
            Pavail = value(Pavail)
        return Pavail - Pused

    def set_initial_condition(
        self, power=None, status=None, hoursinstatus=None
    ):  # pragma: no cover
        self.initial_power = 0
        self.initial_status = 1
        self.initial_status_hours = 0

    def getstatus(self, tend, times=None, status=None):
        return dict(status=1, power=self.power(tend), hoursinstatus=0)

    def create_variables(self, times):
        if self.shedding_mode:
            self.create_variables_shedding(times)
        self.add_parameter(
            "power",
            index=times.set,
            values=dict([(t, self.get_scheduled_ouput(t)) for t in times]),
        )
        self.create_bids(times)

    def create_bids(self, times):
        self.bids = bidding.Bid(
            polynomial=self.cost_coeffs,
            owner=self,
            times=times,
            fixed_input=True,
            status_variable=lambda *args: True,
            input_variable=self.power,
        )

    def create_variables_shedding(self, times):
        self.add_variable("power_used", index=times.set, low=0)

    def create_constraints(self, times):
        if self.shedding_mode:
            for time in times:
                self.add_constraint(
                    "max_power", time, self.power(time) <= self.power_available(time)
                )

    def cost(self, time, scenario=None, evaluate=False):
        return self.operatingcost(time, scenario=scenario, evaluate=evaluate)

    def operatingcost(self, time=None, scenario=None, evaluate=False):
        return self.bids.output(
            time, scenario=scenario
        ) + user_config.cost_wind_shedding * self.shed(
            time, scenario=scenario, evaluate=evaluate
        )

    def truecost(self, time, scenario=None):
        return self.cost(time)

    def incrementalcost(self, time, scenario=None):
        return self.bids.output_incremental(self.power(time))

    def cost_startup(self, time, scenario=None):
        return 0

    def cost_shutdown(self, time, scenario=None):
        return 0

    def cost_first_stage(self, times):
        return 0

    def cost_second_stage(self, times):
        return sum(self.cost(time) for time in times)

    def get_scheduled_ouput(self, time):
        return float(self.schedule.loc[time])

    def set_power_to_observed(self, times):
        power = self.power_available()
        for time in times:
            power[time] = self.observed_values[time]


class Generator_Stochastic(Generator_nonControllable):

    """
    A generator with a stochastic power output.
    """

    def __init__(
        self,
        scenario_values=None,
        costcurveequation="0",
        fuelcost=1,
        bid_points=None,
        noloadcost=0,
        mustrun=False,
        faststart=False,
        pmin=0,
        pmax=None,
        name="",
        index=None,
        bus=None,
        kind="wind",
        observed_values=None,
        schedule=None,
        sheddingallowed=False,
        **kwargs
    ):
        update_attributes(self, locals())  # load in inputs
        self.is_controllable = False

        self.is_stochastic = not (
            user_config.perfect_solve or user_config.deterministic_solve
        )
        self.build_cost_model()
        self.init_optimization()
        self.startupcost = 0
        self.shutdowncost = 0
        self.fuelcost = float(fuelcost)
        self.shedding_mode = sheddingallowed and user_config.economic_wind_shed

    def power(self, time, scenario=None):
        return self.get_variable(
            "power_used" if self.shedding_mode else "power",
            time=time,
            scenario=scenario,
            indexed=True,
        )

    def power_available(self, time=None, scenario=None):
        return self.get_variable("power", time=time, scenario=scenario, indexed=True)

    def _get_scenario_values(self, times, s=0):
        # scenario values are structured as a pd.Panel
        # with axes: day, scenario, {prob, [hours]}
        # the panel has items which are dates
        return (
            self.scenario_values[times.Start.date()][list(range(len(times)))]
            .loc[s]
            .dropna()
            .values.tolist()
        )

    def _get_scenario_probabilities(self, times):
        # if any of the scenario values are defined, we want them
        return self.scenario_values[times.Start.date()].dropna(how="all").probability

    def create_variables(self, times):
        if self.shedding_mode:
            self.create_variables_shedding(times)

        if self.is_stochastic:
            # initialize parameter set to first scenario value
            scenario_one = self._get_scenario_values(times, s=0)
            self.add_parameter(
                "power", index=times.set, values=dict(list(zip(times, scenario_one)))
            )
        else:
            # set to forecast values
            self.add_parameter(
                "power",
                index=times.set,
                values=dict([(t, self.get_scheduled_ouput(t)) for t in times]),
            )

        self.create_bids(times)
        return



from minpower.config import user_config, parse_command_line_config
from minpower.get_data import _load_raw_data, _parse_raw_data, setup_times
from minpower.commonscripts import joindir, set_trace
from minpower.powersystems import PowerSystem
from minpower.solve import create_solve_problem
import pandas as pd
import argparse


def initial_dispatch(directory=".", output_filename="initial.csv"):
    user_config.directory = directory
    user_config.dispatch_decommit_allowed = True

    generators_data, loads_data, lines_data, init_data = _load_raw_data()

    init_data = pd.DataFrame()

    timeseries, times, generators_data, loads_data = setup_times(
        generators_data, loads_data
    )

    # get rid of a bunch of stuff to make this an ED problem
    generators_data["power"] = None
    loads_data["power"] = None
    for i, snm in generators_data.pop("schedulename").dropna().iterkv():
        generators_data.loc[i, "power"] = timeseries[snm].values[0]

    for i, snm in loads_data.pop("schedulename").dropna().iterkv():
        loads_data.loc[i, "power"] = timeseries[snm].values[0]

    for i, snm in generators_data.pop("observedname").dropna().iterkv():
        generators_data.loc[i, "power"] = timeseries[snm].values[0]

    valid_columns = pd.Index(
        [
            "name",
            "pmin",
            "pmax",
            "fuelcost",
            "heatrateequation",
            "costcurveequation",
            "sheddingallowed",
            "power",
        ]
    )

    generators_data = generators_data[generators_data.columns & valid_columns]

    generators, loads, lines, times, scenario_values, data = _parse_raw_data(
        generators_data, loads_data, lines_data, init_data
    )

    power_system = PowerSystem(generators, loads, lines)
    sln = create_solve_problem(power_system, times)

    # save a csv of the dispatch
    dispatch = pd.DataFrame(
        {
            "power": sln.generators_power.loc[0].values,
            "status": sln.generators_status.loc[0].values,
        },
        index=[gen.name for gen in sln.generators],
    )
    dispatch.loc[dispatch.status == 0, "power"] = 0
    dispatch.index.name = "name"
    dispatch.to_csv(joindir(user_config.directory, output_filename))

    print(dispatch)


def main():
    parser = argparse.ArgumentParser(
        "Initial dispatch tool",
        description="""Get the initial dispatch for a case. Runs a economic
        dispatch for the first hour of a commitment problem, but allows 
        generators to turn off to minimize cost.""",
    )
    parser.add_argument("--output_filename", default="initial.csv")

    # parse all of the standard minpower options
    args = parse_command_line_config(parser)
    initial_dispatch(args["directory"], args["output_filename"])


if __name__ == "__main__":
    main()



# minpower's setup script

# make a new release by using
# seed release --dry-run
# and then
# seed release

from setuptools import setup, find_packages

version_number = open("minpower/__init__.py").read().split('"')[1].rstrip('"')

setup(
    name="minpower",
    version=version_number,
    download_url="https://github.com/adamgreenhall/minpower"
    + "/zipball/v{v}".format(v=version_number),
    entry_points="""
    [console_scripts]
    minpower = minpower.solve:main
    standalone_minpower = minpower.solve:standaloneUC
    scheduler_minpower = minpower.experiments.scheduler_minpower:main
    initial_dispatch = minpower.experiments.get_initial_dispatch:main
    """,
    package_data={
        "minpower.configuration": ["minpower.cfg"],
        "minpower.tests": ["*.csv", "*/*.csv", "*/*/*.csv"],
    },
    install_requires=[
        "pandas>=1.3",
        "pyomo>=6.0",
        "matplotlib>=3.4",
        "xarray>=0.18",
    ],
    tests_require=["nose", "coverage", "objgraph"],
    # it helps to have seed if you are going to make releases
    # but it is not required for setup
    description="power systems optimization made beautiful",
    author="Adam Greenhall",
    author_email="minpower@adamgreenhall.com",
    url="http://adamgreenhall.github.io/minpower",
    packages=find_packages(),
    keywords=["power systems", "optimization"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    long_description="""
power systems tools made beautiful
-----------------------------------------

* Solves ED, OPF, and UC problems.
* Problems can be defined in simple spreadsheets.
* Visualizations are created for the answers.
* Many solvers are supported.


* `Full documentation and tutorials <http://adamgreenhall.github.io/minpower>`_
* `Code base <http://github.com/adamgreenhall/minpower>`_

""",
)



"""
Provide the defaults and configuration for other modules.
`user_config` is treated as a global in minpower.
"""
import os
import sys
from configparser import SafeConfigParser
from .commonscripts import DotDict, joindir

parser = SafeConfigParser()
parser.read(
    [
        # the minpower default set, from the minpower/configuration directory
        joindir(os.path.split(__file__)[0], "configuration/minpower.cfg"),
        # the user's global overrides, from the home directory
        os.path.expanduser("~/minpower.cfg"),
        os.path.expanduser("~/.minpowerrc"),
    ]
)

option_types = dict(
    duals=bool,
    breakpoints=int,
    hours_commitment=int,
    hours_overlap=int,
    cost_load_shedding=float,
    cost_wind_shedding=float,
    economic_wind_shed=bool,
    dispatch_decommit_allowed=bool,
    solver=str,
    mipgap=float,
    solver_time_limit=float,
    reserve_fixed=float,
    reserve_load_fraction=float,
    faststart_resolve=bool,
    visualization=bool,
    logging_level=int,
    problem_file=bool,
    output_prefix=bool,
    debugger=bool,
    scenarios=int,
    deterministic_solve=bool,
    perfect_solve=bool,
    scenarios_directory=str,
    cvar_weight=float,
    cvar_confidence_level=float,
    standalone=bool,
    pid=str,
    standalone_restart=bool,
    wind_forecast_adder=float,
    wind_multiplier=float,
    wind_capacity_factor=float,
    wind_error_multiplier=float,
    load_multiplier=float,
    load_adder=float,
    pmin_multiplier=float,
    ramp_limit_multiplier=float,
    ignore_minhours_constraints=bool,
    ignore_ramping_constraints=bool,
    ignore_pmin_constraints=bool,
    on_complete_script=str,
    file_gens=str,
    file_loads=str,
    file_lines=str,
    file_init=str,
    # HACKs to help out resetting the config in testing
    directory=str,
    store_filename=str,
)


def parser_value(parser, section, key, opt_type):
    meth = "get"
    if opt_type == bool:
        meth += "boolean"
    elif opt_type == str:
        pass
    else:
        meth += str(opt_type).split("'")[1]
    return getattr(parser, meth)(section, key)


def parse_config(parser):
    minpower_config = DotDict(
        dict(
            [
                (k, parser_value(parser, "minpower", k, v))
                for k, v in list(option_types.items())
            ]
        )
    )

    s = "scheduler"
    scheduler_config = DotDict(
        dict(
            email=parser.get(s, "email"),
            hours_limit=parser.getint(s, "hours_limit"),
            memory=parser.getint(s, "memory"),
            scheduler_mode=parser.get(s, "scheduler_mode"),
            verbose=parser.getboolean(s, "verbose"),
        )
    )

    return minpower_config, scheduler_config


user_config, scheduler_config = parse_config(parser)

available_solvers = []
for solver in parser.options("available_solvers"):
    if parser.getboolean("available_solvers", solver):
        available_solvers.append(solver)


def filter_non_defaults(x, defaults):
    return dict(
        (k, v) for k, v in list(x.items()) if (k not in defaults) or (v != defaults[k])
    )


def get_dir_config(directory):
    """
    pull the config from a specific directory.
    this allows post-load updating of the user_config.
    """
    dirparser = SafeConfigParser()
    dirparser.read(
        [
            # the minpower default set, from the minpower/configuration directory
            # need this to set the defaults
            joindir(os.path.split(__file__)[0], "configuration/minpower.cfg"),
            # need the home directory overrides too
            os.path.expanduser("~/minpower.cfg"),
            os.path.expanduser("~/.minpowerrc"),
            # the directory's defaults
            joindir(directory, "minpower.cfg"),
        ]
    )

    new_user_config, new_scheduler_conifg = parse_config(dirparser)
    # return (filter_non_defaults(new_user_config, user_config),
    #    filter_non_defaults(new_scheduler_conifg, scheduler_config))
    return new_user_config, new_scheduler_conifg


def parse_command_line_config(parser, preparsed_args=None):
    # get the directory first

    parser = setup_parser_args(parser)

    directory = parser.parse_args(preparsed_args).directory

    # an extra config parse to make sure that if you are calling a case but not
    # in the case directory, the case config gets used
    new_user_config, new_scheduler_config = get_dir_config(directory)
    user_config.update(new_user_config)
    scheduler_config.update(new_scheduler_config)

    parser.set_defaults(**dict(user_config))

    # figure out the command line arguments
    clargs = vars(parser.parse_args(preparsed_args))

    user_config.update(clargs)

    user_config.update(dict(directory=directory))

    if clargs.pop("show_config"):
        from pprint import pprint

        pprint(user_config)
        sys.exit()

    return clargs


def add_opt(parser, name, shortcut=None, help=""):
    kwds = dict(help=help, type=option_types[name], default=user_config[name])
    callers = ["--{n}".format(n=name)]

    if shortcut:
        callers.append(shortcut)
    if kwds["type"] == bool:
        kwds["action"] = "store_true"
        kwds.pop("type")

    parser.add_argument(*callers, **kwds)


def setup_parser_args(parser):
    parser.add_argument(
        "directory", type=str, help="the direcory of the problem you want to solve"
    )

    add_opt(parser, "solver", "-s", help="the solver name (e.g. cplex, gurobi, glpk)")
    add_opt(parser, "visualization", "-v", help="save a visualization of the solution")

    add_opt(
        parser,
        "breakpoints",
        "-b",
        help="number of breakpoints to use in piecewise linearization of polynomial costs",
    )
    add_opt(
        parser,
        "hours_commitment",
        "-c",
        help="number hours per commitment in a rolling UC (exclusive of overlap)",
    )
    add_opt(
        parser,
        "hours_overlap",
        "-o",
        help="number hours to overlap commitments in a rolling UC",
    )

    solver_opt = parser.add_argument_group("Solver options")
    add_opt(solver_opt, "mipgap", help="the MIP gap solution tolerence")
    add_opt(
        solver_opt, "solver_time_limit", help="the MIP solver time limit (in seconds)"
    )

    reserve = parser.add_argument_group(
        "Reserve", "Does the system require reserve? The default is no reserve."
    )
    add_opt(
        reserve,
        "reserve_fixed",
        help="the static amount of reserve required at all times (in MW)",
    )
    add_opt(
        reserve,
        "reserve_load_fraction",
        help="fraction of the total system load which is required as reserve",
    )

    add_opt(
        parser,
        "duals",
        "-d",
        help="flag to get the duals, or prices, of the optimization problem",
    )
    add_opt(
        parser,
        "dispatch_decommit_allowed",
        help="flag to allow de-commitment of units in an ED -- useful for getting initial conditions for UCs",
    )

    add_opt(
        parser,
        "cost_load_shedding",
        help="the cost to the system to shed a MWh of load",
    )
    add_opt(
        parser,
        "cost_wind_shedding",
        help="the cost to the system to shed a MWh of wind energy",
    )
    add_opt(
        parser,
        "economic_wind_shed",
        help="is wind allowed to be shed for economic reasons "
        + "(default is to allow wind shedding only if infeasible)",
    )

    stochastic = parser.add_argument_group(
        "Stochastic UC", "options to modify the behavior of a stochastic problem"
    )
    add_opt(stochastic, "scenarios", help="limit the number of scenarios to N")
    add_opt(
        stochastic,
        "faststart_resolve",
        "-F",
        help="""allow faststart units which are off to be
                started up during resolve with observed wind values""",
    )

    stochastic_mode = stochastic.add_mutually_exclusive_group()
    add_opt(
        stochastic_mode,
        "deterministic_solve",
        "-D",
        help="solve a stochastic problem deterministically using the forecast_filename paramter",
    )
    add_opt(
        stochastic_mode,
        "perfect_solve",
        "-P",
        help="solve a stochastic problem with perfect information",
    )
    add_opt(
        stochastic_mode,
        "scenarios_directory",
        help="override scenarios directory for stochastic problem",
    )

    add_opt(
        stochastic,
        "cvar_weight",
        help="weighting term for CVaR. The default 0 -- this corresponds to an expected value objective function",
    )
    add_opt(
        stochastic,
        "cvar_confidence_level",
        help="confidence level term for a CVaR objective formulation",
    )

    add_opt(
        parser,
        "standalone",
        "-m",
        help="Make each multi-day commitment its own subprocess (helps with memory issues).",
    )
    add_opt(
        parser,
        "output_prefix",
        "-p",
        help="Prefix all results files with the process id (for a record of simulataneous solves)",
    )
    add_opt(parser, "pid", help="Expicitly specify the standalone storage process id")

    debugging = parser.add_argument_group("Debugging tools")
    add_opt(debugging, "debugger", help="use pdb when an error is raised")
    add_opt(
        debugging,
        "problem_file",
        help="flag to write the problem formulation to a problem.lp file",
    )
    add_opt(debugging, "logging_level", help="set the level of detail for logging")
    add_opt(
        debugging,
        "standalone_restart",
        help="restart a multi-stage standalone problem from where it failed",
    )
    debugging.add_argument(
        "--profile",
        action="store_true",
        default=False,
        help="run cProfile and output to minpower.profile",
    )
    debugging.add_argument(
        "--show_config",
        action="store_true",
        default=False,
        help="just show the configuration and quit",
    )
    debugging.add_argument(
        "--keep_lp_files",
        action="store_true",
        default=False,
        help="keep a record of all of the solver lp files",
    )

    ts = parser.add_argument_group(
        "Timeseries modifiers", "Alter the timeseries after parsing the data."
    )

    wind_modifier = ts.add_mutually_exclusive_group()
    add_opt(
        wind_modifier, "wind_multiplier", help="scale the wind power by this factor"
    )
    add_opt(
        wind_modifier,
        "wind_capacity_factor",
        help="scale the wind power to match a given energy capacity factor (defined as a toal wind over total load energy for the timeseries)",
    )

    add_opt(
        ts,
        "wind_error_multiplier",
        help="scale the wind power forecast error by this factor",
    )
    add_opt(
        ts,
        "wind_forecast_adder",
        help="add a fixed amount (e.g. bias) to the wind power forecast",
    )
    add_opt(ts, "load_multiplier", help="scale the load power by this factor")
    add_opt(ts, "load_adder", help="add a fixed amount to the load power")

    constraints = parser.add_argument_group(
        "Ignore/relax constraints",
        "Ignore or relax sets of constraints to allow for feasible solutions.",
    )
    add_opt(
        constraints,
        "ignore_minhours_constraints",
        help="drop the min up/down time constraints on the generators",
    )
    add_opt(
        constraints,
        "ignore_ramping_constraints",
        help="drop the min ramping power constraints on the generators",
    )
    add_opt(
        constraints,
        "ignore_pmin_constraints",
        help="drop the min power constraints on the generators",
    )
    add_opt(
        constraints,
        "pmin_multiplier",
        help="scale the generator minimum power limits by this factor",
    )
    add_opt(
        constraints,
        "ramp_limit_multiplier",
        help="scale the generator ramp power limits by this factor",
    )

    filenames = parser.add_argument_group(
        "Filenames", "Change the default filenames to use when loading data."
    )
    add_opt(filenames, "file_gens")
    add_opt(filenames, "file_loads")
    add_opt(filenames, "file_lines")
    add_opt(filenames, "file_init")

    add_opt(
        parser,
        "on_complete_script",
        help="run a script on completion of the minpower script",
    )

    # NOTE - don't let defaults creep into this defenition
    # that makes resetting the defaults during testing very hard

    return parser



"""
A selection of integration tests. These are designed to find failures that
the unit tests don't pick up.
"""
import os
import nose
from nose.tools import istest
from pandas.testing import assert_frame_equal, assert_series_equal

from minpower.solve import solve_problem as solve_dir
from minpower.config import user_config
from .test_utils import reset_config, with_setup


def basic_checks(sln):
    cost = sln.totalcost_generation
    power = sln.generators_power.loc[cost.index]
    status = sln.generators_status.loc[cost.index]

    # make sure that if status = 0 then power also = 0
    Pcheck = power * status
    assert_frame_equal(Pcheck, power)

    # make sure that if status = 0 then cost also = 0
    Ccheck = cost * status
    assert_frame_equal(Ccheck, cost)

    # make sure that generation meets load (plus any shed)
    shed = sln.load_shed_timeseries
    try:
        scheduled = sln.power_system.total_scheduled_load()
    except AttributeError:
        # look up the scheduled load from the timeseries data
        loads = [
            col for col in sln.store["data_timeseries"].columns if col.startswith("d")
        ]
        scheduled = sln.store["data_timeseries"][loads].sum(axis=1)
    try:
        scheduled.index = shed.index
    except AssertionError:
        # not the same length due to overlap
        scheduled = scheduled.loc[scheduled.index[: len(shed)]]
        scheduled.index = shed.index

    assert_series_equal(scheduled - shed, power.sum(axis=1))


#    except:
#        wind_gen = sln.power_system.get_generator_with_observed()
#        gen_scheduled = pd.DataFrame(dict(
#            wind_obs=wind_gen.observed_values,
#            other_sched=sum(gen.schedule \
#                for gen in sln.power_system.generators() \
#                if not gen.is_controllable and gen != wind_gen)
#            )).loc[:len(shed.index)]
#        gen_scheduled.index = shed.index
#
#        print('controllable_requirement')
#        print(scheduled - gen_scheduled.sum(axis=1))
#        raise


this_directory = os.path.dirname(__file__)
default_config = user_config.copy()


def run_case(name, basedir=this_directory, **kwds):
    """
    a programatic way to do (nearly) the same thing
    as calling the minpower script
    (nearly: because prepends basedir to the directory name you are solving)
    """
    # reset user_config to defaults first
    user_config.update(default_config)
    assert user_config == default_config
    user_config.breakpoints = 3  # for speedier testing
    user_config.update(kwds)
    sln = solve_dir(os.path.join(basedir, name))
    basic_checks(sln)
    return sln


@istest
def run_uc():
    run_case("uc")


@istest
def run_uc_rolling():
    run_case("uc-rolling")


@istest
def run_ed():
    run_case("ed")


@istest
def run_opf():
    run_case("opf")


# just test that the visulizations all work
@istest
@with_setup(teardown=reset_config)
def ed_visualization():
    run_case("ed", visualization=True)


@istest
@with_setup(teardown=reset_config)
def opf_visualization():
    run_case("opf", visualization=True)


@istest
@with_setup(teardown=reset_config)
def uc_visualization():
    duals = user_config.solver != "glpk"
    run_case("uc-rolling", visualization=True, duals=duals)
